

def plot_tsne(docs: list[Document], projections: ndarray = None):
    """
        didnt work?!
    :param docs:
    :param projections:
    :return:
    """
    if projections is None:
        projections = tsne_docs(docs)
    fig = px.scatter(
        projections, x=0, y=1, hover_name=[doc.content for doc in docs],
        color=["black" for doc in docs], labels={'color': 'name'}
        # color=[doc.content for doc in docs], labels={'color': 'name'}
    )
    fig.show()
    plt.show()
    return fig


def tsne_plot_similar_words(title: str, docs: list[Document], embedding_clusters: ndarray):
    """
    didnt work?!
    :param title:
    :param docs:
    :param embedding_clusters:
    :return:
    """
    figsize = (30, 20) if (matplotlib.get_backend() == 'nbAgg') else (20, 12)  # interactive plot should be smaller
    plt.figure(figsize=(figsize))
    # colors = cm.rainbow(np.linspace(0, 1, len(labels)))
    x = embedding_clusters[:, 0]
    y = embedding_clusters[:, 1]
    plt.scatter(x, y, c="red", alpha=0.5, label="")
    for i, doc in enumerate(docs):
        plt.annotate(doc.content, alpha=0.5, xy=(x[i], y[i]), xytext=(5, 2),
                     textcoords='offset points', ha='right', va='bottom', size=8)
    plt.legend(loc=4)
    plt.title(title)
    plt.grid(True)
    # plt.show()


def tsne_html(embeddings_2d: ndarray, labels: list[str], file_name: str):
    fig = go.Figure(data=go.Scattergl(
        x=embeddings_2d[:, 0],
        y=embeddings_2d[:, 1],
        mode='markers',
        text=labels,  # This adds markers' labels
        marker=dict(
            color=np.random.randn(len(labels)),  # set color equal to a variable
            colorscale='Viridis',  # one of plotly colorscales
            line_width=1
        )
    ))

    # Create HTML file
    fig.write_html(file_name)




from enbios2.base.databases import init_databases
from enbios2.ecoinvent.ecoinvent_index import (
    analyze_directory,
    get_ecoinvent_dataset_index,
)

if __name__ == "__main__":
    init_databases()
    analyze_directory(store_to_index_file=True)

    all_indexes = get_ecoinvent_dataset_index()
    for index in all_indexes:
        if index.bw_project_db.exists():
            es = index.bw_project_db.get()
            print(es)
        else:
            if index.version == "3.9.1" and index.type == "default":
                print(f"Should add: {index}")
            print(f"Missing: {index}")



import re
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Optional

from enbios2.generic.files import DataPath


def parse_xml(
    file_path: Path,
    tag: str,
    attributes: list[str],
    value_tag: Optional[list[str]] = None,
) -> list[dict[str, Optional[str]]]:
    def get_namespace(element):
        m = re.match("\{.*\}", element.tag)
        return m.group(0) if m else ""

    text = file_path.read_text(encoding="utf-8")

    root = ET.fromstring(text)
    ns = get_namespace(root)
    result = []

    for elem in root.iterfind(f".//{ns}{tag}"):
        # print(elem)
        data = {attr: elem.attrib.get(attr, None) for attr in attributes}
        # print(elem, elem.attrib, value_tag)
        if value_tag:
            for vt in value_tag:
                value_elem = elem.find(f"{ns}{vt}")
                # print(value_elem)
                if value_elem is not None:
                    data[vt] = value_elem.text
        result.append(data)

    return result


if __name__ == "__main__":
    data_path: DataPath = DataPath(
        "ecoinvent/ecoinvent 3.9.1_cutoff_ecoSpold02/MasterData/Units.xml"
    )
    res = parse_xml(data_path, "unit", [], ["name"])



from typing import Optional

import bw2data
import numpy as np

from bw2data.backends import Activity
from sentence_transformers import SentenceTransformer
from sklearn.manifold import TSNE
from sqlalchemy.orm.attributes import flag_modified
from sqlmodel import Session
from tqdm import tqdm

from enbios2.base.db_models import EcoinventDataset
from enbios2.bw2 import set_bw_current_project
from enbios2.experiment.bw_vector_db.psql_vectorDB import engine, Document, reset_vector_db

set_bw_current_project(EcoinventDataset._SM_CUTOFF, EcoinventDataset._V391)


# print(bw2data.projects.current)


class VectorDBConfig:
    """
    class to store static model
    """
    model = None

    @classmethod
    def get_model(clzz, model_name: str = 'all-MiniLM-L6-v2'):
        """
        load model if not loaded
        :param model_name:
        :return:
        """
        if not clzz.model:
            print("loading model")
            clzz.model = SentenceTransformer(model_name)
        return clzz.model


def insert_db_my_batch(db_name: str, batch_size: int = 1000):
    def process_batch(batch: list[Activity]):
        embeddings = VectorDBConfig.get_model().encode([act["name"] for act in batch])
        with Session(engine) as session:
            for i, embedding in enumerate(embeddings):
                item = Document(content=batch[i]["name"], ext_ids=batch[i]["id"], embedding=embedding.tolist())
                session.add(item)
            session.commit()

    db = bw2data.Database(db_name)

    batch = []
    db_size = len(db)
    for act in tqdm(db, total=db_size):
        act['name']
        # print(act)
        batch.append(act)
        if len(batch) == batch_size:
            process_batch(batch)
            batch = []

    if len(batch) > 0:
        process_batch(batch)


def add_content(content: str, id: int = -1):
    embedding = VectorDBConfig.get_model().encode(content)
    with Session(engine) as session:
        item = Document(content=content, ext_id=id, embedding=embedding.tolist())
        session.add(item)
        session.commit()


def insert_db(db_name: str, normalize: bool = False):
    db = bw2data.Database(db_name)
    contents2ids: dict[str, list[int]] = {}
    for doc in db:
        contents2ids.setdefault(doc["name"], []).append(doc["id"])
    model = VectorDBConfig.get_model()
    contents = list(contents2ids.keys())
    print(f"{len(db)} docs ({len(contents)} names)")
    embeddings = model.encode(contents, show_progress_bar=True)
    if normalize:
        norm_embeddings = VectorDBConfig.get_model().encode(contents, show_progress_bar=True, normalize_embeddings=True)
    with Session(engine) as session:
        for i, embedding in enumerate(embeddings):
            item = Document(content=contents[i], ext_ids=contents2ids[contents[i]],
                            embedding=embedding.tolist())
            if normalize:
                item.norm_embedding = norm_embeddings[i].tolist()
            session.add(item)
        print("committing")
        session.commit()


def query(query_str: str, limit: int = 5):
    """
    make a query to the vector db
    :param query_str:
    :param limit:
    :return:
    """
    embedding = VectorDBConfig.get_model().encode(query_str)
    with Session(engine) as session:
        neighbors = session.query(Document).order_by(
            Document.embedding.cosine_distance(embedding)).limit(limit).all()
        return [(n.content, n.ext_ids) for n in neighbors]


def tsne_docs(docs: list[Document]):
    tsne_ak_2d = TSNE(perplexity=30, n_components=2, init='pca', n_iter=3500, random_state=64, n_jobs=-1)
    embeddings_ak_2d = tsne_ak_2d.fit_transform(np.array(np.array([doc.embedding for doc in docs])))
    return embeddings_ak_2d


def umap_docs(docs: list[Document], use_norm: bool = False, assign_to: Optional[str] = None):
    import umap.umap_ as umap
    reducer = umap.UMAP(random_state=42)
    embeddings = [doc.norm_embedding if use_norm else doc.embedding for doc in docs]
    umap_res = reducer.fit_transform(embeddings)
    if assign_to:
        with Session(engine) as session:
            for i, doc in enumerate(docs):
                setattr(doc, assign_to, umap_res[i])
                flag_modified(doc, assign_to)
                session.add(doc)
            session.commit()
    return umap_res


def get_all_vector_docs() -> list[Document]:
    with Session(engine) as session:
        return session.query(Document).all()

# Call the function with your data
# plot_embeddings(reduced_embeddings, labels)

# uu = umap_docs(get_all(),False, "embedding_2d")
# uu = umap_docs(get_all(),True, "norm_embedding_2d")
# docs = get_all()
# tsne_html(np.array([doc.embedding_2d for doc in docs]),
#           [doc.content for doc in docs], "umap.html")
#
# tsne_html(np.array([doc.norm_embedding_2d for doc in docs]),
#           [doc.content for doc in docs], "umap_norm.html")

# hierarchical_clustering(docs)
# print("done")

if __name__ == "__main__":
    reset_vector_db()
    insert_db("cutoff391")
    while True:
        text = input("Enter text: ")
        res = query(text)
        print(res)



from random import choice

import pandas as pd
from bw2calc import MultiLCA
from bw2data import calculation_setups

import bw2data as bd
from bw2data.backends import Activity


def run_multi_lca(
    name: str, functional_units: dict[Activity:float], impact_methods: list[str]
):
    """
    Perform MultiLCA calculations with many functional units and LCIA methods.

    """
    if len(functional_units) > 0 and len(impact_methods) > 0:
        calculation_setups[name] = {"inv": functional_units, "ia": impact_methods}
        multi_lca = MultiLCA(name)
        index = [str(x) for x in list(multi_lca.all.keys())]
        columns = [str(x) for x in impact_methods]
        results = pd.DataFrame(multi_lca.results, columns=columns, index=index)
        return results
    else:
        raise ValueError("Check the in inputs")


bd.projects.set_current("ecoinvent_391")

db = bd.Database("ecoinvent_391_cutoff")


rand_acts = [{db.random(): 1} for _ in range(2)]
all_methods = list(bd.methods)
methods = list(set(choice(all_methods) for _ in range(1)))

print("activities", rand_acts)
print("methods", methods)

res = run_multi_lca("test", rand_acts, methods)
print(res)

# one demand with 2 funcitonal unit ... short, bad looking code for
# {key1: value1, key2: value2}
# lca = LCA(demand={list(a.keys())[0]: list(a.values())[0] for a in rand_acts},
# method=methods[0], use_distributions=True)
# lca.lci()
# lca.lcia()
#
# # df = pd.DataFrame([{'score': lca.score} for _ in zip(lca, range(10))])
# # print(df)
# print(lca.score)
"""
Also see:
https://stackoverflow.com/questions/42984831/create-a-dataframe-from-multilca-results-in-brightway2
A demand can already include multiple activities!

also read...
https://oie-mines-paristech.github.io/lca_algebraic/example-notebook.html
https://github.com/brightway-lca/from-the-ground-up/blob/main/basic%20tasks/Searching.ipynb
"""




from typing import Optional

from pydantic import BaseSettings


class Settings(BaseSettings):
    huggingface_token: Optional[str] = None

    class Config:
        # `.env.prod` takes priority over `.env`
        env_file = '.env', 'template.env'


settings = Settings()


import datasets
# Print all the available datasets
# print(list_datasets())
# TODO WHERE TO?
# ds = load_dataset("ArmelR/stack-exchange-instruction") # home/.cache/hugging_face
ds = datasets.load_dataset("ArmelR/stack-exchange-instruction", split="train", streaming=True)

#
# login(settings.huggingface_token)
# checkpoint = "bigcode/starcoder"
# device = "cpu"  # for GPU usage or "cpu" for CPU usage
#
# tokenizer = AutoTokenizer.from_pretrained(checkpoint)
# # to save memory consider using fp16 or bf16 by specifying torch.dtype=torch.float16 for example
# model = AutoModelForCausalLM.from_pretrained(checkpoint).to(device)
#
# inputs = tokenizer.encode("def print_hello_world():", return_tensors="pt").to(device)
# outputs = model.generate(inputs)
# print(tokenizer.decode(outputs[0]))



from enbios2.base.experiment import Experiment
import bw2data

from enbios2.models.experiment_models import ExperimentData


bw2data.projects.set_current("ecoinvent")
database_name = "cutoff_3.9.1_default"
db = bw2data.Database(database_name)

wind_turbines_spain = db.search(
    "electricity production, wind, 1-3MW turbine, onshore", filter={"location": "ES"}
)
print(wind_turbines_spain)


experiment_activities = []
for activity in wind_turbines_spain:
    experiment_activities.append(
        {
            "id": {
                "name": activity["name"],
                "location": activity["location"],
                "code": activity["code"],
            }
        }
    )


experiment_activities[0]["output"] = ["kilowatt_hour", 3]
print(experiment_activities)

# select 2 random methods and convert them into the form for enbios2
methods = [bw2data.methods.random() for _ in range(2)]
experiment_methods = [{"id": method} for method in methods]


# let's store the raw data, because we want to modify it later
raw_data = {
    "bw_project": "ecoinvent",
    "activities": experiment_activities,
    "methods": experiment_methods,
}

# make a first validation of the experiment data
exp_data = ExperimentData(**raw_data)

exp: Experiment = Experiment(exp_data)


# run all scenarios at once
results = exp.run()

exp.results_to_csv("test.csv")
# exp.scenarios[0].result_to_dict()


# solar_spain = db.search("solar", filter={"location": "ES"})[:2]
# raw_data["activities"].extend([
#     {
#         "id": {
#             "name": activity["name"],
#             "code": activity["code"]
#         }
#     }
#     for activity in solar_spain
# ])
#
# #%%
# hierarchy = {
#     "wind": [wind_act["name"] for wind_act in wind_turbines_spain],
#     "solar": [solar_act["name"] for solar_act in solar_spain]
# }
#
# raw_data["hierarchy"] = hierarchy
# hierarchy
#
# exp: Experiment = Experiment(raw_data)



import json
from typing import Any, Iterable, Union

import bw2data as bd
from bw2data.backends import Exchange, Activity, ActivityDataset, ExchangeDataset


def system_overview() -> list[tuple]:
    _current = bd.projects.current
    data = []

    names = sorted([x.name for x in bd.projects])
    for obj in names:
        bd.projects.set_current(obj, update=False, writable=False)
        data.append([bd.projects.current] + [[k, v, len(bd.Database(k))] for k, v in bd.databases.items()])
    bd.projects.set_current(_current)
    return data


def graph(database: str) -> Any:
    db = bd.Database(database)
    result = {"nodes": [], "edges": []}
    nodes = result["nodes"]
    edges = result["edges"]

    for obj in db._get_queryset().dicts():
        data = obj["data"]
        nodes.append(
            {"key": obj["code"], "attributes": {**{"size": 15, "label": data["name"], "x": 100, "y": 100}, **{}}})

    from bw2data.backends.schema import ExchangeDataset

    edges_q = ExchangeDataset.select(ExchangeDataset.id, ExchangeDataset.data).where(
        (ExchangeDataset.output_database == database) & (ExchangeDataset.input_database == database)).limit(100)

    for edge in edges_q:
        edges.append({"key": edge.id, "source": edge.data["output"][1], "target": edge.data["input"][1]
                         , "attributes": {"size": 1}})

    return result


def grap_nodes(activities: Iterable[Union[Activity, ActivityDataset]],
               exchanges: Iterable[Union[Exchange, ExchangeDataset]],
               level_infos: Iterable[int]) -> Any:
    result = {"nodes": [], "edges": []}
    nodes = result["nodes"]
    edges = result["edges"]
    levels_iter = iter(level_infos)

    activity_codes: list[str] = []

    current_level = -1
    node_on_current_level = 0
    for obj in activities:
        if isinstance(obj, ActivityDataset):
            dataset_obj: ActivityDataset = obj
        else:
            dataset_obj: ActivityDataset = obj._document
        activity_codes.append(dataset_obj.code)
        level = next(levels_iter)
        if level == current_level:
            node_on_current_level += 1
        else:
            current_level = level
            node_on_current_level = 0
        # print(dataset_obj.name, level, node_on_current_level)
        nodes.append(
            {"key": str(len(activity_codes) - 1),
             "attributes": {**{"size": 15,
                               "label": dataset_obj.name,
                               "color": "#00ff00" if "bio" in dataset_obj.database else "a0a0a0",
                               "x": 100 - level * 30,
                               "y": 100 + node_on_current_level * 30}, **{}}})

    for exc in exchanges:
        if isinstance(exc, ExchangeDataset):
            exc_obj: ExchangeDataset = exc
        else:
            exc_obj: ExchangeDataset = exc._document

        if exc_obj.output_code in activity_codes and exc_obj.input_code in activity_codes:
            edges.append({"key": str(exc_obj.id),
                          "source": str(activity_codes.index(exc_obj.output_code)),
                          "target": str(activity_codes.index(exc_obj.input_code)),
                          "type":"arrow",
                          "attributes": {"size": 1}})

    return result


if __name__ == "__main__":
    # print(json.dumps(system_overview(), indent=2))
    db_name = "cutoff391"
    bd.projects.set_current("ecoi_dbs")
    graph_data = graph(db_name)
    json_data = json.dumps(graph_data, indent=2)
    # open("test.json", "w", encoding="utf-8").write(json.dumps({"nodes": activities, "edges": []}, indent=2))
    # bd.Database("cutoff391").load()



from pint import UnitRegistry

from enbios2.generic.files import DataPath

ureg = UnitRegistry()

ecoinvent_units_file_path = DataPath("ecoinvent_pint_unit_match.txt")

if not ecoinvent_units_file_path.exists():
    print(
        f"Creating 'ecoinvent_pint_unit_match' file at: "
        f"{ecoinvent_units_file_path.as_posix()}"
    )
    ecoinvent_units_file_path.touch()

ureg.load_definitions(ecoinvent_units_file_path)
# print("loaded ecoinvent units")



from os import PathLike
from typing import Union, Optional

from flatten_dict import unflatten
from frictionless import Schema, Resource, validate, system
from frictionless.fields import NumberField, StringField


from enbios2.generic.files import ReadPath
from enbios2.models.experiment_models import (
    ExperimentData,
    ActivitiesDataRows,
    ExperimentMethodData,
)
from enbios2.generic.tree.csv2dict import csv_tree2dict

activities_schema = Schema(
    fields=[
        StringField(name="alias"),
        StringField(name="database"),
        StringField(name="code"),
        StringField(name="name"),
        StringField(name="location"),
        StringField(name="unit"),
        NumberField(name="output.value", float_number=True),
        StringField(name="output.unit"),
    ]
)

activities_unflatten_map = {
    "alias": "id.alias",
    "database": "id.database",
    "code": "id.code",
    "name": "id.name",
    "location": "id.location",
    "unit": "id.unit",
    "output.value": "output.value",
    "output.unit": "output.unit",
}

methods_schema = Schema(
    fields=[
        StringField(name="alias"),
        StringField(name="id"),
        StringField(name="id0"),
        StringField(name="id1"),
        StringField(name="id2"),
        StringField(name="id3"),
    ]
)


def unflatten_data(data: dict, structure_map: dict):
    res = {}
    for key, value in data.items():
        if key in structure_map:
            res[structure_map[key]] = value
        else:
            res[key] = value
    print(res)
    return unflatten(res, splitter="dot")


def parse_method_row(row: dict) -> ExperimentMethodData:
    id_: list[str] = []
    if row_id := row.get("id"):
        id_ = row_id.split(",")
    else:
        for i in range(4):
            if row_id := row.get(f"id{i}"):
                id_.append(row_id)
            else:
                break
    return ExperimentMethodData(alias=row.get("alias"), id=tuple(id_))


def get_abs_path(path: Union[str, PathLike], base_dir: Optional[str] = None) -> ReadPath:
    if base_dir:
        return ReadPath(base_dir) / path
    else:
        return ReadPath(path)


def resolve_input_files(raw_input: ExperimentData):
    # activities
    if isinstance(raw_input.activities, str) or isinstance(
        raw_input.activities, PathLike
    ):
        activities_file: ReadPath = get_abs_path(
            raw_input.activities, raw_input.config.base_directory
        )
        if activities_file.suffix == ".json":
            data = activities_file.read_data()
            raw_input.activities = data
        elif activities_file.suffix == ".csv":
            with system.use_context(trusted=True):
                resource: Resource = Resource(
                    path=activities_file.as_posix(), schema=activities_schema
                )
                report = validate(resource)
                if not report.valid:
                    raise Exception(
                        f"Invalid activities file: {raw_input.activities}; "
                        f"errors: {report.task.errors}"
                    )
                raw_input.activities = ActivitiesDataRows(
                    list(
                        (
                            unflatten_data(r, activities_unflatten_map)
                            for r in resource.read_rows()
                        )
                    )
                )

    # methods
    if isinstance(raw_input.methods, str) or isinstance(raw_input.methods, PathLike):
        methods_file: ReadPath = get_abs_path(
            raw_input.methods, raw_input.config.base_directory
        )
        if methods_file.suffix == ".json":
            data = methods_file.read_data()
            raw_input.methods = data
        if methods_file.suffix == ".csv":
            with system.use_context(trusted=True):
                resource: Resource = Resource(
                    path=methods_file.as_posix(), schema=methods_schema
                )
                report = validate(resource)
                if not report.valid:
                    raise Exception(
                        f"Invalid methods file: {raw_input.methods}; "
                        f"errors: {report.task.errors}"
                    )
                raw_input.methods = list(
                    (parse_method_row(r) for r in resource.read_rows())
                )

    # hierarchy
    if isinstance(raw_input.hierarchy, str) or isinstance(raw_input.hierarchy, PathLike):
        hierarchy_file: ReadPath = get_abs_path(
            raw_input.hierarchy, raw_input.config.base_directory
        )
        if hierarchy_file.suffix == ".json":
            data = hierarchy_file.read_data()
            raw_input.hierarchy = data
        elif hierarchy_file.suffix == ".csv":
            data = csv_tree2dict(hierarchy_file, False)
            raw_input.hierarchy = data
        else:
            raise Exception(f"Invalid hierarchy file: {raw_input.hierarchy}")

    # scenarios
    if isinstance(raw_input.scenarios, str) or isinstance(raw_input.scenarios, PathLike):
        scenario_file: ReadPath = get_abs_path(
            raw_input.scenarios, raw_input.config.base_directory
        )
        if scenario_file.suffix == ".json":
            data = scenario_file.read_data()
            raw_input.scenario = data
        else:
            raise Exception(f"Invalid scenario file: {raw_input.scenarios}")



"""
this is from bw2data sqlite, but is not in all version of bw2data
"""
import json
import pickle
from pathlib import Path

from peewee import BlobField, SqliteDatabase, TextField


class PickleField(BlobField):
    def db_value(self, value):
        return super(PickleField, self).db_value(pickle.dumps(value, protocol=4))

    def python_value(self, value):
        return pickle.loads(bytes(value))


class SubstitutableDatabase:
    def __init__(self, filepath, tables):
        self._filepath = filepath
        self._tables = tables
        self._database = self._create_database()

    def _create_database(self):
        db = SqliteDatabase(self._filepath)
        for model in self._tables:
            model.bind(db, bind_refs=False, bind_backrefs=False)
        db.connect()
        db.create_tables(self._tables)
        return db

    @property
    def db(self):
        return self._database

    def change_path(self, filepath):
        self.db.close()
        self._filepath = filepath
        self._database = self._create_database()

    def atomic(self):
        return self.db.atomic()

    def execute_sql(self, *args, **kwargs):
        return self.db.execute_sql(*args, **kwargs)

    def transaction(self):
        return self.db.transaction()

    def vacuum(self):
        print("Vacuuming database ")
        self.execute_sql("VACUUM;")


class JSONField(TextField):
    """Simpler JSON field that doesn't support advanced querying and is human-readable"""

    def db_value(self, value):
        return super().db_value(
            json.dumps(
                value,
                ensure_ascii=False,
                indent=2,
                default=lambda x: x.isoformat() if hasattr(x, "isoformat") else x,
            )
        )

    def python_value(self, value):
        return json.loads(value)


class TupleJSONField(JSONField):
    def python_value(self, value):
        data = json.loads(value)
        if isinstance(data, list):
            data = tuple(data)
        return data


class PathField(TextField):
    def db_value(self, value):
        if value is not None:
            return str(value)

    def python_value(self, value):
        if value is not None:
            return Path(value)



import functools
from pathlib import Path
from typing import Optional, Any

from enbios2.experiment.tree_transformer import tree_to_csv
from enbios2.generic.tree.basic_tree import BasicTreeNode

@deprecated(reason="Just use BasicTreeNode with type-parameter")
class HierarchyNode(BasicTreeNode):
    """

    Attributes:
    ----------
    name : str
        The name of the node.

    children : list["HierarchyNode"]
        The child nodes of this node.

    value : float, optional
        The value associated with this node.

    parent : HierarchyNode, optional
        The parent node of this node. None if this node is the root.
    """

    def __init__(self,
                 name: str,
                 children: Optional[list["HierarchyNode"]] = (),
                 value: Optional[float] = None,
                 allow_resetting_value: bool = True):
        """
        Initialize the HierarchyNode.

        :param name: The name of the node.
        :param children: A list of child nodes (default is an empty list).
        :param value: The value associated with the node (default is None).
        :param allow_resetting_value: Whether to allow to set the value multiple times
        """
        # call super constructor
        super().__init__(name, children)
        self._value: float = value
        self.children: list["HierarchyNode"] = []  # Override children type
        for child in children:
            self.add_child(child)
        self.parent: Optional[HierarchyNode] = None
        self.allow_resetting_value = allow_resetting_value

    # def add_child(self, node: "HierarchyNode"):  # Override add_child method
    #     self.children.append(node)
    #     node.parent = self

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value: float):
        if self.is_value_set and not self.allow_resetting_value:
            raise ValueError("Value already set")
        self._value = value

    @property
    def is_value_set(self):
        return self.value is not None

    def calc(self, ignore_missing_values: bool = False):
        """
        Calculate the value of the current node. If it has children,
        the value is the sum of the values of the children.

        :param ignore_missing_values: Whether to ignore missing values when
                                      calculating (default is False).
        :return: The calculated value.
        """

        if (not self.children) and (not self.is_value_set) and (not ignore_missing_values):
            raise ValueError(
                f"Hierarchy node '{self.name}': {self.location_names()} must either have children or have its value set")
        if self.value:
            return self.value
        elif self.children:
            self.value = functools.reduce(lambda total, node: total + node.calc(ignore_missing_values),
                                          self.children, 0.0)
            return self.value
        else:  # must be ignore_missing_values = True
            return 0

    def as_dict(self) -> dict[str, Any]:
        """
        Convert the hierarchy from this node down into a dictionary.

        :return: The dictionary representing the hierarchy.
        """

        def rec_build_dict(node: HierarchyNode) -> dict:
            return {
                "name": node.name,
                "children": {
                    child.name: rec_build_dict(child)
                    for child in node.children
                }, "value": node.value}

        return {self.name: rec_build_dict(self)}

    def to_csv(self, file_path: Path, **kwargs):
        """
        Write the hierarchy to a csv file.

        :param file_path: The path to the csv file.
        """
        include_attrs = {"value"}
        if "include_attrs" in kwargs:
            include_attrs.update(kwargs["include_attrs"])
            del kwargs["include_attrs"]
        tree_to_csv(self.as_dict()[self.name], file_path, list(include_attrs), **kwargs)




from enbios2.generic.files import ReadDataPath


def get_ecoinvent_geo_data(file_path: ReadDataPath) -> list[dict]:
    """
    :param file_path:
    :return:
    """
    return [
        {
            "name": geo["name"]["#text"],
            "code": geo["shortname"]["#text"],
            "longitude": geo.get("@longitude"),
            "latitude": geo.get("@latitude"),
        }
        for geo in file_path.read_data()["validGeographies"]["geography"]
    ]


def geo_code2name(ecoinvent_geo_data: list[dict]) -> dict[str, str]:
    return {geo["code"]: geo["name"] for geo in ecoinvent_geo_data}


# a = get_ecoinvent_geo_data(
# ReadDataPath("ecoinvent/ecoinvent 3.9.1_cutoff_ecoSpold02/MasterData/Geographies.xml"))
# code2name

# TODO next step... geoquery...
# from peewee import *
# from playhouse.sqlite_ext import SqliteExtDatabase
#
# # Create a database instance
# db = SqliteExtDatabase('my_database.sqlite', extensions=['mod_spatialite'])
#
# class BaseModel(Model):
#     class Meta:
#         database = db
#
# class TestGeom(BaseModel):
#     id = AutoField()
#     name = CharField(unique=True)
#     geom = TextField()  # Store WKT here
#
# # Connect to the database
# db.connect()
#
# # Create tables
# db.create_tables([TestGeom])
#
# # Insert data into the table
# TestGeom.create(name='Test Point', geom='POINT(10 20)')
#
# # Perform a spatial query
# query = (TestGeom
#          .select()
#          .where(fn.ST_Distance(
#              fn.GeomFromText(TestGeom.geom, 4326),
#              fn.GeomFromText('POINT(10 20)', 4326)
#          ) < 10))
#
# # Fetch and print the results
# for row in query:
#     print(row.name, row.geom)
#
# # Close the connection
# db.close()



"""
A shortcut of initializing the project in brightway2.
The motivation is to have a local brightway project index file.
Now if someone uses this module to set the current index,
that is independent of the machine that is used, as long as the index file exists.
This helps with the problem of different bw project names on different machines
"""
from typing import Optional

import bw2data
import yaml
from bw2data.backends import SQLiteBackend

from enbios2.base.databases import init_databases
from enbios2.base.db_models import BWProjectIndex, EcoinventDataset
from enbios2.ecoinvent.ecoinvent_index import get_ecoinvent_dataset_index
from enbios2.generic.enbios2_logging import get_logger

projects = bw2data.projects

logger = get_logger(__name__)


def _read_bw_index_file() -> list[BWProjectIndex]:
    return list(BWProjectIndex.select())


def print_bw_index():
    """
    Prints the index file
    :return:
    """
    print(_read_bw_index_file())


def get_existing(project_name: str, database_name: str) -> Optional[BWProjectIndex]:
    """
    Get an existing index based on the project and database name
    :param project_name:
    :param database_name:
    :return:
    """
    existing = list(
        BWProjectIndex.select(BWProjectIndex, EcoinventDataset)
        .join(EcoinventDataset)
        .where(
            (BWProjectIndex.project_name == project_name)
            & (BWProjectIndex.database_name == database_name)
        )
    )
    if existing:
        return existing[0]
    return None


def set_bw_index(
    project_name: str, database_name: str, ecoinvent_dataset: EcoinventDataset
) -> BWProjectIndex:
    """
    set a new index
    :param project_name:
    :param database_name:

    :return:
    """
    existing = get_existing(project_name, database_name)
    if existing:
        logger.info(
            f"Index for {project_name},{database_name} exists already with "
            f"Ecoinvent dataset: {existing.ecoinvent_dataset.identity}"
        )
        return existing

    bw_project_index = BWProjectIndex.create(
        project_name=project_name,
        database_name=database_name,
        ecoinvent_dataset=ecoinvent_dataset,
    )
    return bw_project_index


def project_index_creation_helper():
    """
    A helper to get an overview of the projects and databases (should go somewhere else)
    :return:
    """
    projects_overview = {}
    for project in projects:
        bw2data.projects.set_current(project.name)
        projects_overview[project.name] = {
            k: {key: value for key, value in v.items() if key in ["format", "number"]}
            for k, v in bw2data.databases.data.items()
        }
    print(yaml.dump(projects_overview))


def set_bw_current_project(system_model: str, version: str) -> Optional[BWProjectIndex]:
    """

    :param system_model:
    :param version:
    :return: the BWProjectIndex, if the project was set
    """
    bwp = (
        BWProjectIndex.select()
        .join(
            EcoinventDataset,
            on=(BWProjectIndex.ecoinvent_dataset == EcoinventDataset.id),
        )
        .where(
            (EcoinventDataset.system_model == system_model)
            & (EcoinventDataset.version == version)
            & (EcoinventDataset.type == "default")
            & (EcoinventDataset.xlsx is False)
        )
    )
    if bwp:
        bwp = bwp[0]
        bw2data.projects.set_current(bwp.project_name)
        logger.info(
            f"Set current project to '{bwp.project_name}'. "
            f"Ecoinvent database is '{bwp.database_name}'"
        )
        return bwp
    else:
        logger.error(
            f"No brightway project found for ecoinvent dataset: {system_model}, {version}"
        )
        return None


def get_bw_database(system_model: str, version: str) -> Optional[SQLiteBackend]:
    bwp = set_bw_current_project(system_model, version)
    if bwp:
        return bw2data.Database(bwp.database_name)
    return None


if __name__ == "__main__":
    init_databases()
    project_index_creation_helper()
    # print_bw_index()
    candidates = list(
        get_ecoinvent_dataset_index(version="3.9.1", system_model="cutoff", xlsx=False)
    )
    # print(list(candidates))
    set_bw_index("ecoi_dbs", "cutoff391", candidates[0])


#
# def add_bw_project_index(project_name: str):
#     assert project_name in bw2data.projects



from scipy.cluster.hierarchy import linkage, fcluster
from sklearn.cluster import KMeans

from enbios2.experiment.bw_vector_db.create_vectors import get_all_vector_docs
from enbios2.experiment.bw_vector_db.psql_vectorDB import Document


def kmeans(docs: list[Document], k: int = 35):
    kmeans = KMeans(n_clusters=k, random_state=0).fit([doc.embedding for doc in docs])
    return kmeans

def hierarchical_clustering(docs: list[Document]):
    Z = linkage([doc.embedding for doc in docs], method='ward')
    # Plotting dendrogram
    # plt.figure(figsize=(25, 10))
    # plt.title('Hierarchical Clustering Dendrogram')
    # plt.xlabel('sample index')
    # plt.ylabel('distance')
    # dendrogram(
    #     Z,
    #     leaf_rotation=90.,  # rotates the x axis labels
    #     leaf_font_size=8.,  # font size for the x axis labels
    # )
    # plt.show()
    # return Z
    import plotly.figure_factory as ff
    fig = ff.create_dendrogram(Z)
    fig.update_layout(width=800, height=500)
    fig.write_html("dendogram.html")
    # fig.show()
    return Z


def hierarchical(Z):
    max_d = 1  # max_d as in max_distance
    clusters = fcluster(Z, max_d, criterion='distance')

    # Extract keywords for each cluster
    # vectorizer = TfidfVectorizer()
    # unique_clusters = set(clusters)
    # for i in unique_clusters:
    #     cluster_sentences = [sent for sent, cluster in zip(sentences, clusters) if cluster == i]
    #     X = vectorizer.fit_transform(cluster_sentences)
    #     keywords = vectorizer.get_feature_names_out()[X.sum(axis=0).argmax()]
    #     print(f"Cluster {i} keywords: {keywords}")
    return clusters


if __name__ == "__main__":
    print("Running clustering.py")
    docs = get_all_vector_docs()
    k500 = kmeans(docs, 500)



import pytest

# from enbios2.test.test_bw2i import clean_project


def pytest_configure(config):
    config.option.capture = "no"


@pytest.hookimpl(tryfirst=True)
def pytest_sessionfinish(session, exitstatus):
    print("Cleaning up test data...")
    # would delte the project and cause data to be loaded again
    # clean_project()



from pathlib import Path
import os

# enbios2/enbios2/const.py
PROJECT_PATH = Path(os.getcwd())
BASE_DATA_PATH = PROJECT_PATH / "data"
BASE_ECOINVENT_DATASETS_PATH = BASE_DATA_PATH / "ecoinvent"
BASE_DATABASES_PATH = BASE_DATA_PATH / "databases"
MAIN_DATABASE_PATH = BASE_DATABASES_PATH / "meta.sqlite"
BASE_SCHEMA_PATH = BASE_DATA_PATH / "schema"

BASE_TEST_DATA_PATH = BASE_DATA_PATH / "test_data"

# Important data paths/files
BW_PROJECT_INDEX_FILE = BASE_DATA_PATH / "bw_project_index.json"



import csv
from pathlib import Path

from enbios2.generic.files import PathLike


def csv_tree2dict(csv_file: PathLike, ignore_header: bool = True) -> dict:
    tree: dict = {}
    reader = csv.reader(Path(csv_file).open())
    current_node = tree
    _path: list[str] = []

    def insert(_cell: str, _current_node: dict):
        _content = _cell.strip()
        _next_node: dict = {}
        _current_node[_content] = _next_node
        _current_node = _next_node
        _path.append(_content)
        return _current_node

    if ignore_header:
        next(reader)

    for row in reader:
        for depth, cell in enumerate(row):
            if not cell:
                continue
            if cell:
                if not _path:
                    content = cell.strip()
                    current_node = {}
                    tree[content] = current_node
                    _path.append(content)
                else:
                    if depth == len(_path):
                        current_node = insert(cell, current_node)
                    elif depth < len(_path):
                        _path = _path[:depth]
                        current_node = tree
                        for p in _path:
                            current_node = current_node[p]
                        current_node = insert(cell, current_node)
                    else:
                        raise IndexError(
                            f"row: {row} has a cell: '{cell}'"
                            f" that is not positioned correctly"
                        )
    return tree



import csv
from base64 import b64encode
from copy import deepcopy, copy
from pathlib import Path
from typing import Optional, Any, Literal, Union, Generator, TypeVar, Generic, Callable
from uuid import uuid4

from enbios2.generic.enbios2_logging import get_logger
from enbios2.generic.files import PathLike

T = TypeVar("T")

logger = get_logger(__name__)


class BasicTreeNode(Generic[T]):
    """
    The `BasicTreeNode` class represents a node in a tree hierarchy. Each node has a name,
    can have zero or more children.

    Attributes:
    ----------
    name : str
        The name of the node.

    children : list["HierarchyNode"]
        The child nodes of this node.

    parent : HierarchyNode, optional
        The parent node of this node. None if this node is the root.
    """

    def __init__(
        self,
        name: str,
        children: Optional[list[Union["BasicTreeNode[T]", dict[str, Any]]]] = None,
        data: Optional[T] = None,
        data_factory: Optional[Callable[["BasicTreeNode"], T]] = None,
        temp_data: Optional[dict[str, Any]] = None,
    ):
        """
        Initialize the HierarchyNode.

        :param name: The name of the node.
        :param children: A list of child nodes (default is an empty list).
        """
        if not name:
            raise ValueError("name must be a non-empty string")

        self._name: str = name
        self.children: list[BasicTreeNode[T]] = []
        if children:
            for child in children:
                if isinstance(child, dict):
                    child = BasicTreeNode.from_dict(child, data_factory=data_factory)
                self.add_child(child)
        self.parent: Optional[BasicTreeNode[T]] = None
        self.temp_data: dict[str, Any] = temp_data if temp_data else {}
        self._id: bytes = self.generate_id()
        if data:
            self._data: T = data
        elif data_factory:
            self._data: T = data_factory(self)
        else:
            self._data = None

    def generate_id(self) -> bytes:
        self._id = b64encode(uuid4().bytes)
        return self._id

    def set_data(self, data: T):
        self._data = data

    @property
    def data(self) -> T:
        assert self._data is not None, f"data shoould be set before (node: '{self.name}')"
        return self._data

    @property
    def is_leaf(self):
        """
        Check whether this node is a leaf node (i.e., it has no children).

        :return: True if this node is a leaf node, False otherwise.
        """
        return not self.children

    @property
    def level(self) -> int:
        """
        Get the level of this node in the tree. The root node has level 0.
        :return: level of this node
        """
        return len(self.location()) - 1

    @property
    def name(self) -> str:
        """
        Get the name of this node.

        :return: The name of this node.
        """
        return self._name

    @name.setter
    def name(self, name: str):
        """
        Set the name of this node.

        :param name: The new name of this node.
        """
        if self._name == name:
            return
        if self.parent:
            if name in self.parent:
                raise ValueError(f"Node {name} already exists in {self.parent}")
        self._name = name

    def add_child(self, node: "BasicTreeNode") -> "BasicTreeNode":
        """
        Add a child node to this node.

        :param node: The node to be added as a child.
        :return: The node that was added.
        """
        if node is self or node in self:
            raise ValueError(f"Node {node} is already a child of {self}")
        if node.parent:
            raise ValueError(f"Node {node} already has a parent")
        self.children.append(node)
        node.parent = self
        return node

    def add_children(self, nodes: list["BasicTreeNode"]):
        """
        Add multiple child nodes to this node.
        :param nodes:
        :return:
        """
        for node in nodes:
            self.add_child(node)

    def remove_child(self, node: Union["BasicTreeNode[T]", int, str]) -> "BasicTreeNode":
        """
        Remove a child node from this node.

        :param node: The node to be removed.
        :return: The node that was removed.
        """
        if isinstance(node, int) or isinstance(node, str):
            node = self[node]
        if not isinstance(node, BasicTreeNode):
            raise ValueError(f"Node {node} is of wrong type {type(node)}")
        self.children.remove(node)
        node.parent = None
        return node

    def as_dict(
        self,
        include_data: bool = False,
        remove_empty_childlist: bool = False,
        data_serializer: Optional[Callable[[T], Any]] = None,
    ) -> dict[str, Any]:
        """
        Convert the hierarchy from this node down into a dictionary.

        :return: The dictionary representing the hierarchy.
        """
        result = {
            "name": self.name,
            "children": [
                child.as_dict(
                    include_data,
                    remove_empty_childlist=remove_empty_childlist,
                    data_serializer=data_serializer,
                )
                for child in self.children
            ],
        }
        if len(self) == 0 and remove_empty_childlist:
            del result["children"]
        if include_data and self._data:
            result["data"] = (
                data_serializer(self._data) if data_serializer else self._data
            )
        return result

    @staticmethod
    def from_dict(
        data: dict, *, compact: bool = False, data_factory: Optional[Callable] = None
    ) -> "BasicTreeNode":
        """
        Parse a dict and create a tree from it.
        :param data:
        :param compact: if True, the data is assumed to be in compact format
        :param data_factory:
        :return: a node containing the whole tree
        """
        # TODO type parameter of BasicTreeNode
        if compact:
            return BasicTreeNode.from_compact_dict(data, data_factory=data_factory)
        node = BasicTreeNode(**data, data_factory=data_factory)
        return node

    @staticmethod
    def from_compact_dict(
        input_dict, root_name="root", data_factory: Optional[Callable] = None
    ) -> "BasicTreeNode":
        def generate_node(node_info: Union[dict, list]):
            if isinstance(node_info, dict):
                data = node_info.get("data")
                name = node_info.get("name")
                if (
                    not name
                ):  # If 'name' is not a key, then node_info represents children nodes
                    return [
                        BasicTreeNode(
                            name=key,
                            children=generate_node(value),
                            data_factory=data_factory,
                        )
                        for key, value in node_info.items()
                    ]
                else:  # If 'name' is a key, then node_info
                    # represents a single node with optional 'data' and 'children'
                    return [
                        BasicTreeNode(
                            name=name,
                            data=data,
                            children=generate_node(node_info.get("children", [])),
                            data_factory=data_factory,
                        )
                    ]
            elif isinstance(
                node_info, list
            ):  # A list represents a list of children nodes
                nodes = []
                for item in node_info:
                    if isinstance(item, dict):
                        nodes.extend(
                            generate_node(item)
                        )  # A dictionary in a list represents a single node
                        # or multiple children nodes
                    else:  # A string in a list represents a single node
                        nodes.append(BasicTreeNode(name=item, data_factory=data_factory))
                return nodes

        root = BasicTreeNode(
            name=root_name,
            children=generate_node(input_dict),
            data_factory=data_factory,
        )
        return root

    def location(self) -> list["BasicTreeNode[T]"]:
        """
        Get the path from the root to this node.

        :return: List of nodes representing the path from the root to this node.
        """
        nodes: list[BasicTreeNode[T]] = []
        current = self
        while True:
            nodes.append(current)
            if current.parent is None:
                break
            else:
                current = current.parent
        return list(reversed(nodes))

    def location_names(self) -> list[str]:
        """
        Get the names of the nodes in the path from the root to this node.

        :return: List of names representing the path from the root to this node.
        """
        nodes: list[str] = []
        current = self
        while current:
            nodes.append(current.name)
            if not current.parent:
                break
            current = current.parent
        return list(reversed(nodes))

    def location_id(self) -> str:
        """
        Get the ids of the nodes in the path from the root to this node.

        :return: List of ids representing the path from the root to this node.
        """
        return "_".join(self.location_names())

    def assert_all_names_unique(self) -> Literal[True]:
        """
        Assert that all names in the tree are unique. Throws an exception
        if any name is found more than once.
        """
        all_names: set = set()

        def rec_assert_unique_name(node: BasicTreeNode):
            if node.name in all_names:
                print(node.name)
            assert node.name not in all_names, f"{node.name} seems to appear twice"
            all_names.add(node.name)
            for child in node.children:
                rec_assert_unique_name(child)

        rec_assert_unique_name(self)
        return True

    def make_names_unique(self, strategy: Literal["parent_name"] = "parent_name"):
        """
        Make all names in the tree unique.
        The strategy parameter determines how this is done.
        Currently, there is just one strategy,
        which is to prefix the name with the name of the parent
        :param strategy:
        """

        if strategy == "parent_name":
            name_map: dict[str, list[BasicTreeNode[T]]] = {}
            for node in self.iter_all_nodes():
                node.temp_data["orig_name"] = node.name
                name_map.setdefault(node.name, []).append(node)

            for name, nodes in name_map.items():
                if len(nodes) == 1:
                    continue
                parent_level: int = 1

                while len(set((n.name for n in nodes))) != len(nodes):
                    for node in nodes:
                        # in theory this should not happen
                        # if parent_level > node.level:
                        #     raise ValueError(
                        #  f"Cannot make names unique for {name}
                        #  because it is not unique at level {parent_level}")
                        new_name = node.name
                        p = node
                        for i in range(parent_level):
                            assert p.parent is not None
                            p = p.parent
                            new_name = f"{p.temp_data['orig_name']}_{new_name}"
                        node.name = new_name
                    parent_level += 1

    def join_tree(self, node: "BasicTreeNode") -> "BasicTreeNode":
        """
        Add the children of the given node to this  node.

        :param node: The node whose children are to be added.
        :return: This node.
        """
        for child in node.children:
            node.remove_child(child)
            self.add_child(child)
        return self

    def find_subnode_by_name(
        self, name: str, recursive: bool = True
    ) -> Optional["BasicTreeNode"]:
        """
        Find a child node by its name.

        :param name: The name of the node to be found.
        :param recursive: Whether to search recursively into the children's children
                          and so on (default is True).
        :return: The node if found, None otherwise.
        """

        def rec_find_child(node: BasicTreeNode[T]) -> Optional["BasicTreeNode[T]"]:
            if node.name == name:
                return node
            for child in node.children:
                found_node = rec_find_child(child)
                if found_node:
                    return found_node
            return None

        if not recursive:
            for node in [self] + self.children:
                if node.name == name:
                    return node
            return None

        return rec_find_child(self)

    def iter_leaves(self) -> Generator["BasicTreeNode[T]", None, None]:
        """
        Get all leaf nodes of this node.

        :return: List of all leaf nodes.
        """

        def rec_get_leaves(
            node: BasicTreeNode,
        ) -> Generator["BasicTreeNode", None, None]:
            if not node.children:
                yield node
            else:
                for _child in node.children:
                    yield from rec_get_leaves(_child)

        for child in self.children:
            yield from rec_get_leaves(child)

    @property
    def depth(self) -> int:
        """
        Get the depth of this node in the tree.
        Go down the tree until the deepest leaf is reached.
        :return: The depth of this node.
        """

        def calc_max_depth(node):
            if not node.children:
                return 1
            else:
                return 1 + max(
                    (calc_max_depth(child) for child in node.children), default=0
                )

        return calc_max_depth(self)

    def to_csv(
        self,
        csv_file: PathLike,
        *,
        include_data: Optional[bool] = False,
        data_serializer: Optional[Callable[[T], dict]] = None,
        exclude_data_keys: Optional[list[str]] = None,
        level_names: Optional[list[str]] = None,
        merge_first_sub_row: bool = False,
        repeat_parent_name: bool = False,
    ):
        # Calculate max_depth based on root if not provided
        if include_data and not isinstance(self._data, dict) and not data_serializer:
            raise ValueError(
                "If include_data is True, and data not a dict, "
                "data_serializer must be provided"
            )

        if include_data and merge_first_sub_row:
            logger.warning(
                "Merging first sub-row and including data is often not recommended, "
                "as sub-row data will overwrite parent data"
            )

        include_data_keys = []
        if include_data and self._data:
            if data_serializer:
                include_data_keys = list(data_serializer(self._data).keys())
            else:
                if isinstance(self._data, dict):
                    include_data_keys = list(self._data.keys())
            if exclude_data_keys:
                include_data_keys = list(set(include_data_keys) - set(exclude_data_keys))
        _total_level_names = level_names if level_names else []

        def level_name(level: int) -> str:
            if level >= len(_total_level_names):
                _total_level_names.append(f"lvl_{level}")
            return _total_level_names[level]

        def rec_add_node_row(
            node: "BasicTreeNode[T]",
            include_data_: Optional[bool] = False,
            current_level: int = 0,
        ) -> list[dict[str, Union[str, float]]]:
            row = {}
            if include_data_ and node._data:
                node_data: dict[str, Any] = {}
                if data_serializer:
                    node_data = data_serializer(node._data)
                elif isinstance(node._data, dict):
                    node_data = node._data
                else:
                    logger.warning(
                        "Data is not a dict and no data_serializer provided, "
                        "skipping data"
                    )
                for data_key in include_data_keys:
                    row[data_key] = node_data.get(data_key, "")
            row[level_name(current_level)] = node.name
            _sub_rows = []
            for child in node.children:
                _sub_rows.extend(
                    rec_add_node_row(child, include_data_, current_level + 1)
                )
            if _sub_rows:
                if merge_first_sub_row:
                    row = {**row, **_sub_rows[0]}
                    _sub_rows = _sub_rows[1:]
                if repeat_parent_name:
                    for sub_row in _sub_rows:
                        sub_row[level_name(current_level)] = node.name
            return [row] + _sub_rows

        # Write rows to csv
        if isinstance(csv_file, bytes):
            csv_file = csv_file.decode()
        with Path(csv_file).open("w", newline="") as csvfile:
            rows = rec_add_node_row(self, include_data)
            headers = _total_level_names + include_data_keys
            writer = csv.DictWriter(csvfile, headers)
            writer.writeheader()
            writer.writerows(rows)

    def to_sanky_tree(self, file_path: Path, value_key: str = "value"):
        """
        Write the hierarchy to a csv file.
        :param file_path: The path to the csv file.
        :param value_key: The key to use for the value of the links.
        """

        def rec_add_link_row(node, writer: csv.DictWriter, level: int = 0):
            for child in node.children:
                writer.writerow(
                    {
                        "source": node.name,
                        "target": child.name,
                        "value": getattr(child, value_key, 0),
                        "target_level": level + 1,
                    }
                )
                rec_add_link_row(child, writer, level + 1)

        with file_path.open("w", encoding="utf-8") as fout:
            writer = csv.DictWriter(
                fout, fieldnames=["source", "target", "value", "target_level"]
            )
            writer.writeheader()
            rec_add_link_row(self, writer)

    def get_child_names(self) -> list[str]:
        """
        Get the names of all children of this node.
        :return: The names of all children of this node.
        """
        return [child.name for child in self.children]

    def get_num_children(self) -> int:
        """
        Get the number of children of this node.
        :return: The number of children of this node.  == len(self)
        """
        return len(self)

    def collect_all_nodes_at_level(self, level: int) -> list["BasicTreeNode[T]"]:
        """
        Collect all nodes at a given level.

        :param level: The level of the nodes to be collected.
        :return: List of nodes at the given level.
        """
        current = [self]
        for _ in range(level):
            current = [child for node in current for child in node.children]
        return current

    def get_sub_tree(self, max_level: int) -> "BasicTreeNode[T]":
        """
        Get a subtree of the current tree.
        Creates a copy of the tree.

        :param max_level: The maximum level of the subtree.
        :return: The subtree.
        """

        def rec_get_sub_tree(
            node: BasicTreeNode[T], max_level: int, **kwargs
        ) -> BasicTreeNode[T]:
            if max_level == 0:
                return BasicTreeNode[T](
                    node.name, **{k: getattr(node, k) for k in kwargs}
                )
            else:
                sub_tree = BasicTreeNode[T](node.name)
                for child in node.children:
                    sub_tree.add_child(child.get_sub_tree(max_level - 1))
                return sub_tree

        return rec_get_sub_tree(self, max_level)

    def iter_all_nodes(self) -> Generator["BasicTreeNode[T]", None, None]:
        """
        iterates all nodes of the tree. (Depth-First Search)
        :return: Generator of all nodes.
        """
        for _child in self.children:
            yield from _child.iter_all_nodes()

        yield self

    def level_up(self, levels: int) -> "BasicTreeNode":
        """
        Level up the tree by a given number of levels.
        :param levels: The number of levels to level up.
        :return: Returns the node at the new level.
        """
        current = self
        levels_orig = levels
        while levels > 0:
            if not current.parent:
                raise ValueError(
                    f"Cannot level up from {self.name}/level: {self.level}. "
                    f"Requested level: {levels_orig}"
                )
            current = current.parent
            levels -= 1
        return current

    def copy(self, new_name: Optional[str] = None) -> "BasicTreeNode":
        """
        Clone this node (and all children).
        :return: The cloned node.
        """

        _copy = copy(self)
        if new_name:
            _copy.name = new_name
        return _copy

    def __copy__(self) -> "BasicTreeNode":
        """
        Copy the node.
        :return:
        """

        def reset_parents(node: BasicTreeNode):
            node.generate_id()
            for child in node.children:
                child.parent = node
                reset_parents(child)

        node = deepcopy(self)
        reset_parents(node)
        return node

    def copy_an_merge(
        self, child_names: list[str], parent_name: Optional[str] = None
    ) -> "BasicTreeNode[T]":
        """
        Copy this node and all children given in  child_names.
        :param child_names: The names of the children to be copied.
        :param parent_name: The name of the new root node. Default: self.name
        :return: node that contains all "copies" as children.
        """
        _root = BasicTreeNode[T](parent_name if parent_name else self.name)
        for child_name in child_names:
            _root.add_child(self.copy(child_name))
        return _root

    @staticmethod
    def from_csv(
        csv_file: Path,
        node_columns: Optional[list[str]] = None,
        merged_first_sub_row: Optional[bool] = True,
    ) -> "BasicTreeNode[T]":
        reader: csv.DictReader = csv.DictReader(csv_file.open("r", encoding="utf-8"))
        if not node_columns:
            if reader.fieldnames:
                node_columns = list(reader.fieldnames)
            else:
                raise ValueError("No fieldnames in csv.")

        root: Optional["BasicTreeNode[T]"] = None
        current_node: Optional["BasicTreeNode[T]"] = None
        for row in reader:
            for index, lvl_name in enumerate(node_columns):
                if not row[lvl_name] or (clean_value := row[lvl_name].strip()) == "":
                    continue
                if index == 0 and not root:
                    root = BasicTreeNode[T](clean_value)
                    current_node = root
                    continue
                if current_node:
                    if current_node.level < index - 1:
                        raise ValueError(
                            f"There are nodes missing inbetween "
                            f"{current_node.name} and {clean_value}"
                        )
                    elif current_node.level >= index:
                        current_node = current_node.level_up(
                            current_node.level - index + 1
                        )
                    current_node = current_node.add_child(BasicTreeNode(clean_value))

        if not root:
            raise ValueError("Could not generate tree from csv.")
        else:
            return root

    def recursive_apply_eager(
        self,
        func: Callable[["BasicTreeNode", Any], Any],
        depth_first: bool = False,
        *args,
        **kwargs,
    ):
        if not depth_first:
            func(self, *args, **kwargs)

        for child in self.children:
            child.recursive_apply_eager(func, depth_first, *args, **kwargs)

        if depth_first:
            func(self, *args, **kwargs)

    def recursive_apply_lazy(
        self,
        func: Callable[["BasicTreeNode", Any], Any],
        depth_first: bool = False,
        *args,
        **kwargs,
    ) -> Generator[Any, Any, Any]:
        if not depth_first:
            yield func(self, *args, **kwargs)

        for child in self.children:
            for res in child.recursive_apply_lazy(func, depth_first, *args, **kwargs):
                yield res

        if depth_first:
            yield func(self, *args, **kwargs)

    def recursive_apply(
        self,
        func: Callable[["BasicTreeNode", Optional[Any]], Any],
        depth_first: bool = False,
        lazy: bool = False,
        *args,
        **kwargs,
    ):
        """
        note that only when lazy is set to True, the function will return a generator.
        :param func: the function to apply to all nodes
        :param depth_first:
        :param lazy: does not evaluate the function until the generator is iterated.
        :param args:
        :param kwargs:
        :return:
        """
        if lazy:
            return self.recursive_apply_lazy(func, depth_first, *args, **kwargs)
        else:
            self.recursive_apply_eager(func, depth_first, *args, **kwargs)

    def get_child(self, child_index_name: Union[int, str]) -> "BasicTreeNode":
        """
        Get a child node by its index or name.
        Throws KeyError or IndexError if the child is not found.
        :param child_index_name: int index or name of the child node.
        :return: Child node.
        """
        return self[child_index_name]

    def contains(self, item: Union[str, "BasicTreeNode"]) -> bool:
        """
        Check if a given item is a child of this node. calls "item in self"
        :param item: The item or name to check.
        :return: bool, True if the item is a child of this node, False otherwise.
        """
        return item in self

    def clear(self) -> int:
        """
        Remove all children from this node.
        :return: The number of removed children.
        """
        num_children = len(self)
        for _ in range(num_children):
            self.remove_child(0)
        return num_children

    @property
    def id(self):
        return self._id

    def __len__(self):
        """
        Get the number of children of this node.
        :return: The number of children of this node.
        """
        return len(self.children)

    def __getitem__(
        self, item: Union[int, str, list[Union[int, str]]]
    ) -> "BasicTreeNode":
        """
        Get a child node by its index or name.
        Throws KeyError or IndexError if the child is not found.
        :param item: int index or name of the child node.
        :return: Child node.
        """
        if isinstance(item, str):
            for child in self.children:
                if child.name == item:
                    return child
            raise KeyError(f"Node {self.name} has no child with name {item}")
        elif isinstance(item, int):
            return self.children[item]
        elif isinstance(item, list):
            if not item:
                return self
            next_node = self[item[0]]
            return next_node[item[1:]]

    def __contains__(self, item: Union[str, "BasicTreeNode"]) -> bool:
        """
        Check if a given item is a child of this node.
        :param item: The item or name to check.
        :return: bool, True if the item is a child of this node, False otherwise.
        """
        # print(item)
        if isinstance(item, BasicTreeNode) or issubclass(type(item), BasicTreeNode):
            assert not isinstance(item, str)
            item = item.name
        for child_name in self.get_child_names():
            if child_name == item:
                return True
        return False

    def __repr__(self) -> str:
        """
        Get the  string representation of the node

        :return: String representation of the object.
        """
        children_str = f"{len(self.children)} {'children' if len(self) != 1 else 'child'}"
        parent_str = f"{' (' + self.parent.name + ')' if self.parent else ''}"
        return f"[{self.name} - {children_str}{parent_str}]"

    def info(self) -> str:
        if not self._data:
            return repr(self)
        else:
            return repr(self) + f"\n{repr(self._data)}"

    def __bool__(self):
        """
        this prevents that 'if not node' is true for a node that has no children
        :return:
        """
        return True

    def to_mermaid_str(
        self,
        arrow_style="---",
        markdown_context: bool = False,
        leaves_style: str = "fill:#ccc",
    ) -> str:
        """
        Get a string representation of the tree in mermaid format.
        :return: String representation of the tree in mermaid format.
        """

        def safe_name(node: BasicTreeNode):
            return node.name.replace(" ", "_")

        def node_repr(node: BasicTreeNode):
            return f"{safe_name(node)}[{node.name}]"

        def style(node: BasicTreeNode):
            if leaves_style and not node.children:
                return ":::leaf"
            else:
                return ""

        def mm_connection(node: BasicTreeNode[T], _=None):
            return (
                f"{safe_name(node.parent)} {arrow_style} {node_repr(node)}{style(node)}"
                if node.parent
                else ""
            )

        lines = "\n".join([con for con in self.recursive_apply(mm_connection) if con])
        if leaves_style:
            lines += f"\nclassDef leaf {leaves_style}"
        if markdown_context:
            return f"```mermaid\ngraph BT;\n{lines}\n```"
        else:
            return f"graph BT;\n{lines}"



import json
from typing import Tuple, Optional

import bw2data
import networkx
import numpy
import orjson as orjson
from bw2data.backends import ActivityDataset, ExchangeDataset, Activity
from networkx import DiGraph

# print(projects)
# projects.set_current("ecoi_dbs")
# db = Database("cutoff391")

from enbios2.bw2.project_index import print_bw_index, set_bw_current_project
from enbios2.const import BASE_DATA_PATH


def check_unique_codes():
    activity_codes = [a.code for a in ActivityDataset.select(ActivityDataset.code)]
    assert len(activity_codes) == len(set(activity_codes))


def get_tree(code: str,
             keep_exchange_type: list[str] = None,
             check_unique_code: bool = True,
             max_level: int = -1) -> Tuple[set[str], list[int]]:
    """
    Get all nodes that are connected to the given node. (as inputs)
    :param code: root code
    :param keep_exchange_type: keep the exchanges of the given type
    :param check_unique_code: check if the codes are unique
    :param max_level: max level to go down. -1 for infinite
    :return:
    """
    if check_unique_code:
        check_unique_codes()

    # all visited nodes (codes)
    visited: set[str] = set()
    # nodes to visit next
    to_visit = {code}
    # nodes to visit in the next iteration
    to_visit_next = set()
    # all exchanges (eventually filtered by type). just ids for memory efficiency
    all_exchanges: list[int] = []
    current_level: int = 0

    while len(to_visit) and max_level != 0:
        # get all exchanges that we could currently reach
        exchanges = list(
            ExchangeDataset.select(ExchangeDataset.id, ExchangeDataset.input_code, ExchangeDataset.output_code).where(
                ExchangeDataset.output_code.in_(to_visit)))
        # add nodes that we did not visit yet for the next iteration
        for exc in exchanges:
            if exc.input_code == exc.output_code:
                continue
            if exc.input_code not in visited:
                to_visit_next.add(exc.input_code)
        # updated visited nodes
        visited.update(to_visit)
        # save exchanges
        if not keep_exchange_type:
            all_exchanges.extend([e.id for e in exchanges])
        else:
            all_exchanges.extend([exc.id for exc in exchanges if exc.type in keep_exchange_type])
        # update nodes to visit for next iteration
        to_visit = to_visit_next.copy()
        # reset nodes to visit in next iteration
        to_visit_next.clear()

        print(f"visited: {len(visited)}, total exchanges: {len(all_exchanges)}, next nodes: {len(to_visit)}")
        max_level -= 1
        current_level += 1
    return visited, all_exchanges


def get_tree_with_levels(code: str,
                         keep_exchange_type: Optional[list[str]] = None,
                         check_unique_code: Optional[bool] = True,
                         max_level: Optional[int] = -1) -> Tuple[dict[str, int], dict[str, set[str]], list[int]]:
    """
    Get all nodes that are connected to the given node. (as inputs)
    :param code: root code
    :param keep_exchange_type: keep the exchanges of the given type
    :param check_unique_code: check if the codes are unique
    :param max_level: max level to go down. -1 for infinite
    :return:
    """
    if check_unique_code:
        check_unique_codes()

    # all visited nodes (codes)
    visited: set[str] = set[str]()
    # keep a dict of code -> level
    level: dict[str, int] = dict()
    # keep a dict. code -> all input codes
    inputs: dict[str, set[str]] = dict()
    # nodes to visit next
    to_visit = {code}
    # nodes to visit in the next iteration
    # all exchanges (eventually filtered by type). just ids for memory efficiency
    all_exchanges: list[int] = []
    current_level: int = 0

    while len(to_visit) and max_level != 0:
        print("---")
        # get all exchanges that we could currently reach
        visited.update(to_visit)
        for v in to_visit:
            level[v] = current_level
            inputs[v] = set()

        exchanges = list(
            ExchangeDataset.select(ExchangeDataset).where(
                ExchangeDataset.output_code.in_(to_visit)))

        to_visit.clear()
        # add nodes that we did not visit yet for the next iteration
        for exc in exchanges:
            # ignore if production
            if exc.input_code == exc.output_code:
                continue
            if exc.input_code not in visited:
                to_visit.add(exc.input_code)
                inputs[exc.output_code].add(exc.input_code)
            else:
                # get level of visited node
                def rec_inc(node: str, new_level):
                    level[node] = new_level
                    for inp in inputs.get(node, []):
                        rec_inc(inp, new_level + 1)

                rec_inc(exc.input_code, level[exc.output_code] + 1)
        # save exchanges
        if not keep_exchange_type:
            all_exchanges.extend([e.id for e in exchanges])
        else:
            all_exchanges.extend([exc.id for exc in exchanges if exc.type in keep_exchange_type])
        print(f"visited: {len(visited)}, total exchanges: {len(all_exchanges)}, next nodes: {len(to_visit)}")
        max_level -= 1
        current_level += 1
    # sort by the second element of the tuple (level)
    return level, inputs, all_exchanges


if __name__ == "__main__":
    print_bw_index()
    set_bw_current_project("cutoff", "3.9.1")
    print(list(bw2data.databases))
    # random_act = db.random()
    db = bw2data.Database("cutoff_3.9.1_default")
    random_act: Activity = db.get_node("46875148bb5fbac9cad7452d020a80de")
    print(random_act)
    # # print(info_exchanges(random_act))
    #
    calc = True
    # 'non-ionic surfactant production, ethylene oxide derivate' (kilogram, GLO, None)

    if calc:
        levels, inputs, exchanges = get_tree_with_levels(random_act["code"])
        # json.dump((list(visited_with_levels), list(exchanges)), open("temp.json", "w"))
    else:
        visited_with_levels, exchanges = json.load(open("temp.json", "r"))


    # visited, levels = zip(*visited_with_levels)
    # activity_iter = iter_activities_by_codes(iter(visited))

    # graph = grap_nodes(activity_iter, iter_exchange_by_ids(iter(exchanges)), level_infos=levels)
    # json.dumps(graph)

    def default(obj):
        if isinstance(obj, set):
            return list(obj)
        if isinstance(obj, numpy.ndarray):
            return list(obj)
        raise TypeError


    (BASE_DATA_PATH / "fast_tree_data.json").write_text(str(orjson.dumps({
        "levels": levels,
        "inputs": inputs,
        "exchanges": exchanges
    }, default=default)), encoding="utf-8")

    activities = list(ActivityDataset.select().where(ActivityDataset.code.in_(list((levels.keys())))))
    (BASE_DATA_PATH / "fast_tree_activities_data.json").write_text(str(orjson.dumps(
        [[a.key, a.id, a.name, a.location, a.product, a.type, a.data.get("unit")] for a in activities],
        default=default)), encoding="utf-8")

    uuid2id = {a.code: a.id for a in activities}

    (BASE_DATA_PATH / "fast_tree_activities_data_keyed.json").write_text(str(orjson.dumps(
        {str(a.id): [a.database, a.id, a.name, a.location, a.product, a.type, a.data.get("unit"),
                     [uuid2id[i] for i in inputs[a.code]]] for a in activities},
        default=default)), encoding="utf-8")

    graph = DiGraph()
    graph.add_nodes_from([(act.id, {"name": act.name}) for act in activities])
    for act in activities:
        for i in inputs[act.code]:
            graph.add_edge(uuid2id[i], act.id)

    # this will take a long time...
    pos = networkx.spring_layout(graph)
    (BASE_DATA_PATH / "fast_tree_graph_positions.json").write_text(str(orjson.dumps({str(k): v for k, v in pos.items()},
                                                                                    default=default)), encoding="utf-8")



from typing import Generator, Iterator

import bw2data
from bw2data import databases as bw_databases
from bw2data.project import projects as bw_projects

from bw2data.backends import Activity, ExchangeDataset, ActivityDataset


def info_exchanges(act: Activity) -> dict:
    """
    Show exchanges of an activity
    :param act:
    :return:
    """
    exchanges = act.exchanges()
    for exc in exchanges:
        print(exc)
    for exc in act.upstream():
        print(exc)
    return {exc.input: exc for exc in act.exchanges()}


def iter_exchange_by_ids(
    ids: Iterator[int], batch_size: int = 1000
) -> Generator[ExchangeDataset, None, None]:
    """
    Iterate over exchanges by ids
    :param ids:
    :param batch_size:
    :return:
    """
    last_batch: bool = False
    while True:
        batch_ids: list[int] = []
        for i in range(batch_size):
            try:
                batch_ids.append(next(ids))
            except StopIteration:
                last_batch = True
                break
        for exc in ExchangeDataset.select().where(ExchangeDataset.id.in_(batch_ids)):
            yield exc
        if last_batch:
            break


def iter_activities_by_codes(
    codes: Iterator[str], batch_size: int = 1000
) -> Generator[ActivityDataset, None, None]:
    """
    Iterate over activities by codes
    :param codes:
    :param batch_size:
    :return:
    """
    last_batch: bool = False
    while True:
        batch_codes: list[str] = []
        for i in range(batch_size):
            try:
                batch_codes.append(next(codes))
            except StopIteration:
                last_batch = True
                break
        for act in ActivityDataset.select().where(ActivityDataset.code.in_(batch_codes)):
            yield act
        if last_batch:
            break


def get_activity(code: str) -> Activity:
    """
    Get activity by code
    :param code:
    :return:
    """
    activity_ds = ActivityDataset.get_or_none(ActivityDataset.code == code)
    if not activity_ds:
        raise ValueError(f"Activity with code '{code}' does not exist")
    return Activity(activity_ds)


def full_duplicate(activity: Activity, code=None, **kwargs) -> Activity:
    """
    Make a copy of an activity with its upstream exchanges
    (Otherwise, you cannot calculate the lca of the copy)
    :param activity: the activity to copy
    :param code: code of the new activity
    :param kwargs: other data for the copy
    :return: new activity
    """
    activity_copy = activity.copy(code, **kwargs)
    for upstream in activity.upstream():
        upstream.output.new_exchange(
            input=activity_copy, type=upstream["type"], amount=upstream.amount
        ).save()
    activity_copy.save()
    return activity_copy


def clean_delete(activity: Activity):
    """
    Delete an activity and its upstream exchanges.
    Otherwise, the system will be corrupted
    :param activity: The activity to delete
    """
    for link in activity.upstream():
        link.delete()
    activity.delete()


def report():
    current_ = bw2data.projects.current
    projects = list(bw2data.projects)
    for project in projects:
        print(project)
        bw_projects.set_current(project.name)
        databases = list(bw_databases)
        print(databases)
    bw2data.projects.set_current(current_)


def bw_unit_fix(unit_str: str):
    if unit_str == "kilowatt hour":
        return "kilowatt_hour"
    if unit_str == "unit":
        return "unspecificEcoinventUnit"
    return unit_str


def delete_all_projects():
    resp = input("Are you sure you want to delete all projects? [y]")
    if resp != "y":
        print("cancelled")
        return
    for project in bw2data.projects:
        if project.name != "default":
            bw2data.projects.delete_project(project.name, True)
            print(f"Deleted {project.name}")
    bw2data.projects.purge_deleted_directories()


if __name__ == "__main__":
    report()
    # bw2data.projects.purge_deleted_directories()



import bw2data

from enbios2.ecoinvent import get_ecoinvent_dataset_index
from enbios2.base.db_models import EcoinventDataset
from enbios2.bw2 import set_bw_current_project


print(get_ecoinvent_dataset_index())

set_bw_current_project(EcoinventDataset._V391, EcoinventDataset._SM_CUTOFF)
# bw2data.DataStore
print(bw2data.databases)


import bw2data
import bw2io
from bw2io import ExcelImporter

from enbios2.bw2.util import report
from enbios2.generic.files import DataPath

report()
project_name = "excel_import"

reset = False

if reset:
    bw2data.projects.delete_project(project_name, True)
    bw2data.projects.purge_deleted_directories()

if project_name not in bw2data.projects:
    bw2data.projects.create_project(project_name)
    bw2data.projects.set_current(project_name)
    bw2io.bw2setup()
else:
    bw2data.projects.set_current(project_name)

file_path = DataPath("test_data/bw2/csv_excel/a.xlsx")
imported = ExcelImporter(file_path)
imported.apply_strategies()
if imported.all_linked:
    imported.write_database()

print(len(bw2data.Database("csv_db")))
act = list(bw2data.Database("csv_db"))[0]
print(act)
print(act.temp_data)


import json
import os
from csv import DictReader
from pathlib import Path
from typing import Generator, Union, Optional

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader
import openpyxl
import xmltodict as xmltodict

import yaml

from enbios2.const import BASE_DATA_PATH
from enbios2.generic.enbios2_logging import get_logger

logger = get_logger(__name__)


def read_data(path: Path, config: Optional[dict] = None):
    """
    Read data from file. Formats supported: json, csv, excel
    - json is read straight into a dict
    - csv is read into a DictReader object
    - excel is read into a dict of sheet names and lists of rows

    :return:
    """
    if path.suffix == ".json":
        return json.loads(path.read_text(encoding="utf-8"))
    elif path.suffix == ".csv":
        if not config:
            config = {}
        return list(DictReader(path.open(encoding="utf-8"), **config))
    # excel
    elif path.suffix == ".xlsx":
        workbook = openpyxl.load_workbook(path, read_only=True, data_only=True)
        return {sheet.title: list(sheet.values) for sheet in workbook.worksheets}
    # xml
    elif path.suffix == ".xml":
        try:
            import xmltodict
        except ImportError:
            raise ImportError("xmltodict not installed")
        return xmltodict.parse(path.read_text(encoding="utf-8"))
    else:
        raise NotImplementedError(f"File format {path.suffix} not supported")


class DataPath(Path):
    _flavour = Path(".")._flavour  # type: ignore

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, BASE_DATA_PATH, *args, **kwargs)

    def read_data(self, config: Optional[dict] = None):
        """
        Read data from file. Formats supported: json, csv, excel
        - json is read straight into a dict
        - csv is read into a DictReader object
        - excel is read into a dict of sheet names and lists of rows

        :return:
        """
        return read_data(self, config=config)


class ReadPath(Path):
    """
    Checks on instantiation that the file exists
    """

    _flavour = Path(".")._flavour  # type: ignore

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)
        if not instance.exists():
            raise FileNotFoundError(f"File {instance} does not exist")
        return instance

    def read_data(self, config: Optional[dict] = None):
        """
        Read data from file. Formats supported: json, csv, excel
        - json is read straight into a dict
        - csv is read into a DictReader object
        - excel is read into a dict of sheet names and lists of rows

        :return:
        """
        return read_data(self, config=config)

    def iter_data(self) -> Union[dict, Generator[dict, None, Optional[dict]]]:
        if self.suffix == ".json":
            logger.warning("Reading json completely not as iterator")
            return json.loads(self.read_text(encoding="utf-8"))
        elif self.suffix in [".yaml", ".yml"]:
            return yaml.load(self.read_text(encoding="utf-8"), Loader=Loader)
        elif self.suffix == ".csv":
            reader = DictReader(self.open(encoding="utf-8"))
            for row in reader:
                yield row
        elif self.suffix == ".xlsx":
            workbook = openpyxl.load_workbook(self, read_only=True, data_only=True)
            for sheet in workbook.worksheets:
                for row in sheet.iter_rows(values_only=True):
                    yield {sheet.title: row}
        elif self.suffix == ".xml":
            logger.info("Reading xml completely not as iterator")
            return dict(xmltodict.parse(self.read_text(encoding="utf-8")))
        return None


class ReadDataPath(ReadPath):
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, BASE_DATA_PATH, *args, **kwargs)
        if not instance.exists():
            raise FileNotFoundError(f"File {instance} does not exist")
        return instance


PathLike = Union[str, os.PathLike]



import json
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import Optional, Union

import bw2data
from bw2data.backends import ActivityDataset
from prompt_toolkit import PromptSession
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.formatted_text.base import FormattedText

try:
    import enbios2
    # import enbios2.generic
    # from enbios2.generic.enbios2_logging import get_logger
except ImportError:
    print("importing enbios2")
    # print(err)
    import sys

    for dir_ in reversed(Path(__file__).parents):
        if dir_.name == "enbios2":
            sys.path.insert(0, dir_.as_posix())
            break
from enbios2.generic.enbios2_logging import get_logger


class State(Enum):
    PROJECT = "project"
    IN_PROJECT = "in_project"
    DATABASE = "database"
    IN_DATABASE = "in_database"
    METHOD = "method"


logger = get_logger(__name__)


class CustomBuffer(Buffer):
    def insert_completion(self, completion, overwrite_before=0):
        self.delete_before_cursor(overwrite_before)
        self.text = completion.text  # Replace the entire text with the completion


tab_message = ". Press [tab] or [space] to list all options 🌕"
exit_message = ". Press [ctrl+c] to exit 🔚"

GLOBAL_COMMAND_RESET = "RESET"
STORE_DATA = "store"


@dataclass
class ActivityData:
    name: str
    code: str
    database: str
    unit: str


@dataclass
class PromptStore:
    """
    stuff that is stored and collected and will be exported
    """
    project: Optional[str] = None
    methods: list[tuple] = field(default_factory=list)
    activities: list[ActivityData] = field(default_factory=list)  # name, code, database


@dataclass
class PromptState:
    state: State = State.PROJECT
    current_db: bw2data.backends.base.SQLiteBackend = None
    current_candidates: list[str] = field(default_factory=list)  # current candidates depend on the status
    activity_candidates_data: list[ActivityData] = field(default_factory=list)  # used for activities
    selected_candidates: list[str] = field(default_factory=list)  # for multi-selection


class CustomCompleter(Completer):
    def __init__(self):
        super(CustomCompleter, self).__init__()
        self.message: Union[str, FormattedText] = ""  # current message before each input
        self.state: PromptState = PromptState()
        self.data: PromptStore = PromptStore()

        self.reset()

    def reset(self):
        self.data = PromptStore()
        self.state = PromptState()

        self.set_message("select a brightway project")
        self.state.current_candidates = self.get_candidates()

    def get_candidates(self) -> list[str]:
        """
        get the candidates depending on the current status
        :return: all candidates for the current status
        """
        if self.state.state == State.PROJECT:
            return [p.name for p in bw2data.projects]
        elif self.state.state == State.IN_PROJECT:
            return [State.DATABASE.value, State.METHOD.value, STORE_DATA]
        elif self.state.state == State.DATABASE:
            return bw2data.databases

        elif self.state.state == State.METHOD:
            return ["_".join(m) for m in bw2data.methods]
        return []

    def get_coded_candidates(self) -> list[tuple[str, ActivityData]]:
        if self.state.state == State.IN_DATABASE:
            # logger.debug(f"current_db: {self.state.current_db}")
            candidates = list((act.name,
                               ActivityData(name=act.name, code=act.code, unit=act.data["unit"],
                                            database=self.state.current_db.name)) for act in
                               ActivityDataset.select(ActivityDataset.name,
                                                      ActivityDataset.code,
                                                      ActivityDataset.data).where(
                                   ActivityDataset.database == self.state.current_db.name))
            return candidates

    def get_completions(self, document, complete_event):
        word = document.get_word_before_cursor()

        # If there's no word before the cursor, yield all words as completions.
        if not word:
            for known_word in self.state.current_candidates:
                yield Completion(known_word, start_position=-len(document.text))
            return

        # If the word before the cursor is a substring of a known word, yield it as a completion.
        for known_word in self.state.current_candidates:
            if word.lower() in known_word.lower():
                if known_word in self.state.selected_candidates:
                    style = 'bg:#a0ff70'  # green
                    yield Completion(known_word, start_position=-len(document.text), style=style)
                else:
                    yield Completion(known_word, start_position=-len(document.text))

    def set_status(self, new_state: State, text: Optional[str] = None):
        """
        set the status and update the candidates depending on the new status
        :param new_state:
        :return:
        """
        self.state.state = new_state
        if self.state.state == State.IN_PROJECT:
            self.set_message("select 'database', 'method' or 'store' (to store your selections)")
        if self.state.state == State.IN_DATABASE:
            self.state.current_db = bw2data.Database(text)
            # unzip get_coded_candidates into two lists
            self.state.current_candidates, self.state.activity_candidates_data = zip(*self.get_coded_candidates())
        else:
            self.state.current_candidates = self.get_candidates()

    def set_message(self, msg: Union[str, list[tuple[str, str]]]):
        """
        set the message before the input.
        can be string or formatted text. Will add newline and >>> automatically
        :param msg:
        :return:
        """
        if isinstance(msg, str):
            self.message = msg + "\n>>> "
        else:
            self.message = FormattedText(msg + [("", "\n>>>")])

    def back_check(self, text: str, back_state: State) -> bool:
        if text.lower() == "back":
            # self.set_message("select 'database' or 'method'")
            self.set_status(back_state)
            return True
        return False

    def set_next(self, text: str):
        logger.debug(f"{self.state.state} / set_next: {text}")
        if text == GLOBAL_COMMAND_RESET:
            self.reset()
        elif self.state.state == State.PROJECT:
            # print(text, type(text), self.state.current_candidates)
            if text in self.state.current_candidates:
                print("selected project: ", text)
                bw2data.projects.set_current(text)
                self.data.project = text
                self.set_status(State.IN_PROJECT)
            else:
                print(f"project: {text} not found")
        elif self.state.state == State.IN_PROJECT:
            if text == State.DATABASE.value:
                self.set_status(State.DATABASE)
                if not self.state.current_candidates:
                    print("There are no databases...")
                    self.set_status(State.IN_PROJECT)
                else:
                    self.set_message(
                        f"There are {len(self.state.current_candidates)} databases. "
                        f"Select any number of candidates one by one")
            elif text == State.METHOD.value:
                self.set_status(State.METHOD)
                if not self.state.current_candidates:
                    print("There are no methods...")
                    self.set_status(State.IN_PROJECT)
                else:
                    self.set_message(
                        f"There are {len(self.state.current_candidates)} methods. Select any number of "
                        f"candidates one by one")
            elif text == STORE_DATA:
                self.store_data()
            else:
                print(f"command: {text} not found")
        elif self.state.state == State.DATABASE:
            if self.back_check(text, State.IN_PROJECT):
                return
            if text not in bw2data.databases:
                return
            print("loading database...")
            self.set_status(State.IN_DATABASE, text)
            self.set_message(
                f"There are {len(self.state.current_candidates)} activities. Select any number of candidates one by one")
        elif self.state.state == State.IN_DATABASE:
            if self.back_check(text, State.IN_PROJECT):
                return
            try:
                selected_index = self.state.current_candidates.index(text)
                # activity_name = self.state.current_candidates[selected_index]
                activity_data = self.state.activity_candidates_data[selected_index]
                self.data.activities.append(activity_data)
            except ValueError:
                print(f"activity: '{text}' not found")
                return
        elif self.state.state == State.METHOD:
            if self.back_check(text, State.IN_PROJECT):
                return
            try:
                # check validity
                selected_index = self.state.current_candidates.index(text)
                selected_method = list(bw2data.methods)[selected_index]
                if selected_method in self.data.methods:
                    # remove the method again
                    self.data.methods.remove(selected_method)
                    self.state.selected_candidates.remove(text)
                    self.set_message([("", "You "),
                                      ("bg:#a00000", "deselected"),
                                      ("",
                                       f" {text}. You selected {len(self.state.selected_candidates)} methods.")])
                else:
                    self.data.methods.append(selected_method)
                    self.state.selected_candidates.append(text)
                    self.set_message(f"You selected {len(self.state.selected_candidates)} methods.")
            except ValueError:
                print(f"method: {text} not found")

    def toolbar_msg(self):
        messages: list[tuple[str, str]] = []
        if self.state.state == State.PROJECT:
            messages.append(('bold', "🔥 Select a brightway project"))
        elif self.state.state == State.IN_PROJECT:
            messages.append(('bold', f"🔥 Project: '{self.data.project}'. Select 'database' or 'method'"))
        elif self.state.state == State.METHOD:
            messages.append(('bold', f"🔥 Project: '{self.data.project}'. Select methods... (type 'back' to go back)"))
        return FormattedText(messages + [("", tab_message), ("", exit_message)])

    def store_data(self):
        if not self.data.methods:
            print("no methods selected. Not storing")
            return
        if not self.data.activities:
            print("no activities selected. Not storing")
            return

        experiment_data = {
            "bw_project": self.data.project,
            "activities": [
                {"id":
                    {
                        "name": activity.name,
                        "code": activity.code,
                        "database": activity.database,
                    },
                    "output": [
                        activity.unit,
                        1
                    ]} for activity in self.data.activities
            ],
            "methods": [{"id": method} for method in self.data.methods]
        }
        json.dump(experiment_data, open("experiment_data.json", "w"))


if __name__ == "__main__":
    custom_completer = CustomCompleter()

    session = PromptSession(completer=custom_completer,
                            bottom_toolbar=custom_completer.toolbar_msg)

    while True:
        try:
            text = session.prompt(custom_completer.message)
            custom_completer.set_next(text)
        except KeyboardInterrupt:
            text = input("enter 'q' to quit. Anything else to get back to work\n>>> ")
            if text == "q":
                break



from pathlib import Path
from typing import Optional, Union

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from matplotlib.projections import PolarAxes
from pandas import DataFrame

from enbios2.base.result_select import ResultsSelector
from enbios2.base.experiment import Experiment
from enbios2.generic.enbios2_logging import get_logger
from enbios2.generic.files import PathLike
from enbios2.generic.tree.basic_tree import BasicTreeNode
from enbios2.models.experiment_models import ScenarioResultNodeData

logger = get_logger(__name__)


def bar_plot(
    experiment: Union[Experiment, ResultsSelector],
    scenarios: Optional[list[str]] = None,
    methods: Optional[list[str]] = None,
    short_method_names: bool = True,
    image_file: Optional[PathLike] = None,
) -> Figure:
    rs = ResultsSelector.get_result_selector(experiment, scenarios, methods)

    n_rows = len(rs.methods)
    n_cols = 1

    fig, axs = plt.subplots(
        n_rows, n_cols, figsize=(10, 5 * n_rows)
    )  # Assuming each subplot has a height of 5, adjust as needed

    # Check if there's only one subplot to handle the indexing appropriately
    if n_rows == 1:
        axs = [axs]

    for idx, method in enumerate(rs.methods):
        cmap = plt.colormaps.get_cmap("tab10")
        colors = cmap(np.linspace(0, 1, len(rs.scenarios)))
        rs.base_df.plot(kind="bar", x="scenario", y=method, ax=axs[idx], color=colors)
        axs[idx].set_ylabel(rs.method_label_names(short_method_names)[idx], fontsize=8)
        axs[idx].legend().set_visible(False)

    plt.tight_layout()
    if image_file:
        fig.write_image(Path(image_file).as_posix())
    return fig  # Return the figure object


def stacked_bar_plot(
    experiment: Union[Experiment, ResultsSelector],
    scenarios: Optional[list[str]] = None,
    methods: Optional[list[str]] = None,
    level: int = 1,
    short_method_names: bool = True,
    aliases: Optional[list[str]] = None,
    image_file: Optional[PathLike] = None,
) -> Figure:
    rs = ResultsSelector.get_result_selector(experiment, scenarios, methods)
    experiment = rs.experiment

    if level >= experiment.hierarchy_root.depth:
        logger.warning(
            f"Level {level} is higher or equal (>=) than the depth of the hierarchy "
            f"({experiment.hierarchy_root.depth}). "
            f"Limiting to {experiment.hierarchy_root.depth - 1}"
        )
        level = experiment.hierarchy_root.depth - 1
    # Define the number of rows and columns for the subplots
    n_rows = len(rs.methods)
    n_cols = 1

    # Create a new figure with a defined size (adjust as needed)
    # Explicitly create a Figure object
    fig, axs = plt.subplots(
        n_rows, n_cols, figsize=(10, 5 * n_rows)
    )  # Assuming each subplot has a height of 5, adjust as needed

    # Check if there's only one subplot to handle the indexing appropriately
    if n_rows == 1:
        axs = [axs]

    for idx, method in enumerate(rs.methods):
        ax = axs[idx]

        # Create the bar plot using the specific Axes object
        nodes: list[BasicTreeNode[ScenarioResultNodeData]] = []
        if aliases:
            for alias in aliases:
                node = experiment.hierarchy_root.find_subnode_by_name(alias)
                if not node:
                    raise ValueError(f"Alias {alias} not found in hierarchy")
                nodes.append(node)
        else:
            nodes = experiment.hierarchy_root.collect_all_nodes_at_level(level)

        node_names = [node.name for node in nodes]
        df = rs.collect_tech_results(node_names)
        df_pivot = df.pivot(index="scenario", columns="tech", values=method)
        df_pivot.plot(kind="bar", stacked=True, ax=ax)
        ax.set_ylabel(rs.method_label_names(short_method_names)[idx], fontsize=8)

    plt.tight_layout()
    if image_file:
        fig.write_image(Path(image_file).as_posix())
    return fig  # Return the figure object


def star_plot(
    experiment: Union[Experiment, ResultsSelector],
    scenarios: Optional[list[str]] = None,
    methods: Optional[list[str]] = None,
    *,
    fill: bool = True,
    r_ticks=(0.2, 0.4, 0.6, 0.8, 1.0),
    show_r_ticks: bool = True,
    show_grid: bool = True,
    col: int = 4,
    row: Optional[int] = None,
    short_method_names: bool = True,
    show_labels: bool = True,
    image_file: Optional[PathLike] = None,
) -> Figure:
    rs = ResultsSelector.get_result_selector(experiment, scenarios, methods)
    df = rs.normalized_df()

    labels = rs.method_label_names(short_method_names, False)

    if row is None:
        row = int(np.ceil(len(rs.scenarios) / col))
    # if only one row. limit the figure size to the number of scenarios
    if row == 1:
        col = len(rs.scenarios)
    # Create figure and axes
    fig, axs = plt.subplots(
        row, col, figsize=(6 * col, 6 * row), subplot_kw=dict(polar=True)
    )
    if row == 1 and col == 1:
        axs = [[axs]]  # to handle indexing later on
    elif row == 1:
        axs = [axs]
    elif col == 1:
        axs = [[a] for a in axs]

    # remove empty axes
    last_row_num_axes = len(rs.scenarios) % col
    # print(last_row_num_axes)
    if last_row_num_axes != 0:
        for c in range(col - last_row_num_axes):
            # print(c,"--")
            axs[-1][-1 - c].remove()
            # pass

    all_done = False
    for r in range(row):
        for c in range(col):
            ax: PolarAxes = axs[r][c]
            scenario_name: str = rs.scenarios[r * col + c]
            values = df.loc[df["scenario"] == scenario_name].values.tolist()[0][1:]
            values.append(values[0])

            angles = np.linspace(0, 2 * np.pi, len(values) - 1, endpoint=False).tolist()
            angles.append(0)
            # Plot data

            if fill:
                ax.fill(angles, values, color="blue", alpha=0.1)
            else:
                ax.plot(angles, values)

            # Fix axis to close the circle
            ax.set_theta_offset(np.pi / 2)
            ax.set_theta_direction(-1)
            ax.set_rlabel_position(30)
            ax.set_rticks(list(r_ticks))
            if not show_r_ticks:
                ax.set_yticklabels([])
            if not show_grid:
                ax.grid(False)
            angles.pop()
            ax.set_xticks(angles)
            if r == 0 and c == 0 and show_labels:
                ax.set_xticklabels(labels)
            else:
                ax.set_xticklabels([""] * len(angles))
            ax.set_rmax(1)
            ax.title.set_text(scenario_name)

            if r * col + c == len(rs.scenarios) - 1:
                all_done = True
                break
        if all_done:
            break

    plt.tight_layout()
    if image_file:
        fig.write_image(Path(image_file).as_posix())
    return fig


def single_star_plot(
    experiment: Union[Experiment, ResultsSelector],
    scenarios: Optional[list[str]] = None,
    methods: Optional[list[str]] = None,
    *,
    r_ticks=(0.2, 0.4, 0.6, 0.8, 1.0),
    show_r_ticks: bool = True,
    show_grid: bool = True,
    col: int = 4,
    row: Optional[int] = None,
    short_method_names: bool = True,
    show_labels: bool = True,
    image_file: Optional[PathLike] = None,
) -> Figure:
    """
    plots multiple scenarios into one star plot
    :param experiment:
    :param scenarios:
    :param methods:
    :param r_ticks:
    :param show_r_ticks:
    :param show_grid:
    :param col:
    :param row:
    :param short_method_names:
    :param show_labels:
    :param image_file:
    :return:
    """
    rs = ResultsSelector.get_result_selector(experiment, scenarios, methods)
    df = rs.normalized_df()

    labels = rs.method_label_names(short_method_names, False)

    # Create figure and axes
    fig, ax = plt.subplots(figsize=(6, 6), subplot_kw=dict(polar=True))

    cmap = plt.colormaps.get_cmap("tab10")
    for idx, scenario_name in enumerate(scenarios):
        values = df.loc[df["scenario"] == scenario_name].values.tolist()[0][1:]
        values.append(values[0])

        angles = np.linspace(0, 2 * np.pi, len(values) - 1, endpoint=False).tolist()
        angles.append(0)

        _color = list(cmap(idx)[:3]) + [0.3]

        ax.plot(angles, values)

    # Fix axis to close the circle
    ax.set_theta_offset(np.pi / 2)
    ax.set_theta_direction(-1)
    ax.set_rlabel_position(30)
    ax.set_rticks(list(r_ticks))

    if not show_grid:
        ax.grid(False)
    angles.pop()
    ax.set_xticks(angles)
    if show_labels:
        ax.set_xticklabels(labels)
    else:
        ax.set_xticklabels([""] * len(labels))
    ax.set_rmax(1)


def plot_heatmap(
    experiment: Union[Experiment, ResultsSelector],
    scenarios: Optional[list[str]] = None,
    methods: Optional[list[str]] = None,
    special_df: Optional[DataFrame] = None,
    image_file: Optional[PathLike] = None,
) -> Figure:
    rs = ResultsSelector.get_result_selector(experiment, scenarios, methods)
    if special_df is not None:
        rs.check_special_df(special_df)
        df = special_df
    else:
        df = rs.normalized_df()

    df = df.set_index("scenario").transpose()

    fig, ax = plt.subplots(figsize=(len(rs.scenarios) * 1.5, len(rs.methods) * 1.5))
    im = ax.imshow(df, cmap="summer")

    labels = rs.method_label_names(include_unit=False)
    ax.set_xticks(np.arange(len(rs.scenarios)), labels=rs.scenarios)
    ax.set_yticks(np.arange(len(rs.methods)), labels=labels)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")
    ax.figure.colorbar(im, ax=ax, **{})
    for i in range(len(rs.scenarios)):
        for j in range(len(rs.methods)):
            ax.text(
                i,
                j,
                "%.2f" % df[df.columns[i]][j],
                ha="center",
                va="center",
                color="black",
            )

    fig.tight_layout()
    if image_file:
        fig.write_image(Path(image_file).as_posix())
    return fig


def plot_sankey(
    exp: Experiment,
    scenario_name: str,
    method_: str,
    *,
    image_file: Optional[PathLike] = None,
) -> Figure:
    try:  # type: ignore
        import plotly.graph_objects as go
    except ImportError:
        logger.error("plotly not installed. Run 'pip install plotly'")
        return

    scenario = exp.get_scenario(scenario_name)
    all_nodes = list(scenario.result_tree.iter_all_nodes())
    node_labels = [node.name for node in all_nodes]

    source = []
    target = []
    value = []
    for index_, node in enumerate(all_nodes):
        _: BasicTreeNode[ScenarioResultNodeData] = node
        for child in node.children:
            source.append(index_)
            target.append(all_nodes.index(child))
            value.append(child.data.results[method_])

    fig = go.Figure(
        data=[
            go.Sankey(
                node=dict(
                    pad=15,
                    thickness=20,
                    line=dict(color="black", width=0.5),
                    label=node_labels,
                    color="blue",
                ),
                link=dict(source=source, target=target, value=value),
            )
        ]
    )

    fig.update_layout(title_text=f"{scenario.alias} / {method_}", font_size=10)
    if image_file:
        fig.write_image(Path(image_file).as_posix(), width=1800, height=1600)
    return fig


def one_axes_scatter_plot(
    experiment: Union[Experiment, ResultsSelector],
    selected_scenario: str,
    methods: Optional[list[str]] = None,
    image_file: Optional[PathLike] = None,
) -> Figure:
    rs = ResultsSelector.get_result_selector(experiment, None, methods)
    df = rs.normalized_df()

    scenario_index = df[df.columns[0]].tolist().index(selected_scenario)
    n_rows = len(rs.methods)
    fig, axs = plt.subplots(
        len(rs.methods), 1, figsize=(10, 2 * n_rows)
    )  # Assuming each subplot has a height of 5, adjust as needed

    for method_index in range(n_rows):
        ax = axs[method_index]

        x = df[df.columns[method_index + 1]]
        y = np.random.normal(loc=0, scale=1e-286, size=len(df))
        colors = ["#FFA50090"] * len(x)
        colors[scenario_index] = "blue"
        ax.scatter(x, y, s=100, c=colors, marker="o")
        # plt.title("Scatter Plot with Dense Y Distribution")
        ax.set_xlabel(rs.methods[method_index])
        ax.set_yticks([])  # Hide y-axis labels
        ax.grid(True, which="both", linestyle="--", linewidth=0.5, axis="x")
        # no border on all edges
        for spine in ax.spines.values():
            spine.set_visible(False)
    plt.tight_layout()
    if image_file:
        fig.write_image(Path(image_file).as_posix())
    return fig  # Return the figure object



import bw2data

def delete_db(project_name: str, db_name: str):
    if project_name not in bw2data.projects:
        raise ValueError(f"Project '{project_name}'does not exist")
    bw2data.projects.set_current(project_name)
    if db_name not in bw2data.databases:
        raise ValueError(f"Database '{db_name}'does not exist")
    print("proceeding to delete database ...")
    bw2data.Database(db_name).delete_instance()


if __name__ == '__main__':
    delete_db("ecoinvent", "cutoff_3.9.1_default")



import math
import time
from dataclasses import dataclass, field
from datetime import timedelta
from typing import Optional, Union, TYPE_CHECKING, Any

from bw2calc import LCA
from bw2data.backends import Activity
from numpy import ndarray
from pint import DimensionalityError, Quantity, UndefinedUnitError
from pint.facets.plain import PlainQuantity

from enbios2.base.stacked_MultiLCA import StackedMultiLCA
from enbios2.base.unit_registry import ureg
from enbios2.bw2.util import bw_unit_fix
from enbios2.generic.enbios2_logging import get_logger
from enbios2.generic.files import PathLike

# for type hinting
if TYPE_CHECKING:
    from enbios2.base.experiment import Experiment
from enbios2.generic.tree.basic_tree import BasicTreeNode
from enbios2.models.experiment_models import (
    BWCalculationSetup,
    ScenarioResultNodeData,
    ExperimentMethodPrepData,
    Activity_Outputs,
)

logger = get_logger(__name__)


@dataclass
class Scenario:
    experiment: "Experiment"
    alias: str
    result_tree: BasicTreeNode[ScenarioResultNodeData]
    results: Optional[ndarray] = None
    _has_run: bool = False
    # this should be a simpler type - just str: float
    activities_outputs: Activity_Outputs = field(default_factory=dict)
    methods: Optional[dict[str, ExperimentMethodPrepData]] = None
    stored_lcas: list[LCA] = field(default_factory=list)
    _execution_time: float = float("NaN")

    def prepare_tree(self):
        """Prepare the result tree for calculating scenario outputs.
        This populates the result tree with ScenarioResultNodeData objects
        for each activity node, which store the output amount and units.
        If config is set, it also stores the BW activity dict with the node.
        """

        # activity_nodes = list(self.result_tree.get_leaves())
        activities_simple_ids = list(self.activities_outputs.keys())
        for result_index, activity_id in enumerate(activities_simple_ids):
            activity_alias = activity_id.alias
            bw_activity = self.experiment.get_activity(activity_alias).bw_activity
            try:
                activity_node = self.result_tree.find_subnode_by_name(activity_alias)
            except StopIteration:
                raise ValueError(f"Activity {activity_alias} not found in result tree")
            activity_node._data = ScenarioResultNodeData(
                output=(
                    bw_unit_fix(bw_activity["unit"]),
                    self.activities_outputs[activity_id],
                )
            )
            if self.experiment.config.include_bw_activity_in_nodes:
                activity_node._data.bw_activity = bw_activity
        self.result_tree.recursive_apply(
            Scenario._recursive_resolve_outputs,
            depth_first=True,
            scenario=self,
            cancel_parents_of=set(),
        )

    def _create_bw_calculation_setup(self, register: bool = True) -> BWCalculationSetup:
        inventory: list[dict[Activity, float]] = []
        for activity_alias, act_out in self.activities_outputs.items():
            bw_activity = self.experiment.get_activity(activity_alias.alias).bw_activity
            inventory.append({bw_activity: act_out})

        methods = [m.id for m in self._get_methods().values()]
        calculation_setup = BWCalculationSetup(self.alias, inventory, methods)
        if register:
            calculation_setup.register()
        return calculation_setup

    @staticmethod
    def _propagate_results_upwards(
            node: BasicTreeNode[ScenarioResultNodeData], add_to_distribution: bool = False
    ):
        for child in node.children:
            if child.data:
                if add_to_distribution:
                    for key, value in child.data.distribution_results.items():
                        if node.data.distribution_results.get(key) is None:
                            num_distribution = len(
                                list(child.data.distribution_results.values())[0]
                            )
                            node.data.distribution_results[key] = [0] * num_distribution
                        node.data.distribution_results[key] = list(
                            [
                                a + b
                                for a, b in zip(
                                node.data.distribution_results[key],
                                child.data.distribution_results[key],
                            )
                            ]
                        )
                else:
                    for key, value in child.data.results.items():
                        if node.data.results.get(key) is None:
                            node.data.results[key] = 0
                        node.data.results[key] += value

    def _add_lca_results_to_tree(
            self,
            lca_results: ndarray,
            add_to_distribution: bool = False,
            propagate: bool = True,
    ) -> BasicTreeNode[ScenarioResultNodeData]:
        """Add LCA results to each node in the technology tree.

        This takes an array of LCA results and assigns them to each activity
        node in the tree. It loops through activity IDs, gets the node,
        and adds the result for each LCA method to the node's results dict.

        It calls propagate_results_upwards to sum child results in parent nodes.

        Args:
            lca_results: Array of LCA results for each activity and method.

        Returns:
            The updated result tree with all nodes containing results.
        """

        if not self.result_tree:
            raise ValueError(f"Scenario '{self.alias}' has no results...")
        activity_ids = list(self.activities_outputs.keys())

        methods_aliases: list[str] = list(self._get_methods().keys())
        for result_idx, activity_id in enumerate(activity_ids):
            activity_alias = activity_id.alias
            activity_node = self.result_tree.find_subnode_by_name(activity_alias)
            assert activity_node
            for method_idx, method_name in enumerate(methods_aliases):
                if add_to_distribution:
                    node_data = activity_node.data
                    if method_name not in node_data.distribution_results:
                        node_data.distribution_results[method_name] = []
                    node_data.distribution_results[method_name].append(
                        lca_results[result_idx][method_idx]
                    )
                else:
                    activity_node.data.results[method_name] = lca_results[result_idx][
                        method_idx
                    ]

        if propagate:
            self.result_tree.recursive_apply(
                Scenario._propagate_results_upwards,
                depth_first=True,
                add_to_distribution=add_to_distribution,
            )
        return self.result_tree

    def _get_methods(self) -> dict[str, ExperimentMethodPrepData]:
        if self.methods:
            return self.methods
        else:
            return self.experiment.methods

    @staticmethod
    def _recursive_resolve_outputs(
            node: BasicTreeNode[ScenarioResultNodeData], _: Optional[Any] = None, **kwargs
    ):
        # todo, does this takes default values when an activity is not defined
        #  in the scenario?
        scenario: Scenario = kwargs["scenario"]
        cancel_parts_of: set = kwargs["cancel_parents_of"]
        if node.is_leaf:
            return
        node_output: Optional[Union[Quantity, PlainQuantity]] = None
        if any(child.id in cancel_parts_of for child in node.children):
            node.set_data(ScenarioResultNodeData())
            cancel_parts_of.add(node.id)
            return
        for child in node.children:
            # if not child._data:
            #     raise ValueError(f"Node {child.name} has no data")
            activity_output = child.data.output[0]
            if activity_output is None:
                node_output = None
                logger.warning(f"No output unit of activity '{child.name}'.")
                break
            output = None
            try:
                output = ureg.parse_expression(activity_output) * child.data.output[1]
                if not node_output:
                    node_output = output
                else:
                    node_output += output
            except UndefinedUnitError as err:
                logger.error(
                    f"Cannot parse output unit '{activity_output}' of activity "
                    f"{child.name}. {err}. "
                    f"Consider the unit definition to 'enbios2/base/unit_registry.py'"
                )
                node_output = None
                break
            except DimensionalityError as err:
                set_base_unit = node_output.to_base_units() if node_output else ""
                logger.warning(
                    f"Cannot aggregate output to parent: {node.name}. "
                    f"From earlier children the base unit is {set_base_unit} "
                    f"and from {child.name} it is {output}."
                    f" {err}"
                )
                node_output = None
                break
        if node_output:
            node_output = node_output.to_compact()
            node.set_data(
                ScenarioResultNodeData(
                    output=(str(node_output.units), node_output.magnitude)
                )
            )
        else:
            node.set_data(ScenarioResultNodeData())
            logger.warning(
                f"Scenario: '{scenario.alias}': No output for node: '{node.name}' "
                f"(lvl: {node.level}). "
                f"Not calculating any upper nodes."
            )
            cancel_parts_of.add(node.id)

    def run(self) -> BasicTreeNode[ScenarioResultNodeData]:
        if not self._get_methods():
            raise ValueError(f"Scenario '{self.alias}' has no methods")
        logger.info(f"Running scenario '{self.alias}'")
        distributions_config = self.experiment.config.use_k_bw_distributions
        distribution_results = distributions_config > 1
        start_time = time.time()
        bw_calc_setup = self._create_bw_calculation_setup()
        result_tree = {}
        for i in range(distributions_config):
            stacked_lca = StackedMultiLCA(
                bw_calc_setup, distribution_results
            )
            if self.experiment.config.store_lca_objects:
                self.stored_lcas.append(stacked_lca.lca)
            raw_results: ndarray = stacked_lca.results
            result_tree = self.set_results(
                raw_results, distribution_results, i == distributions_config - 1
            )
        self._execution_time = time.time() - start_time
        return result_tree

    @property
    def execution_time(self) -> str:
        if not math.isnan(self._execution_time):
            return str(timedelta(seconds=int(self._execution_time)))
        else:
            return "not run"

    def reset_execution_time(self):
        self._execution_time = float("NaN")

    def set_results(
            self, results: ndarray, add_to_distribution: bool = False, propagate: bool = True
    ) -> BasicTreeNode[ScenarioResultNodeData]:
        if self.experiment.config.store_raw_results:
            self.results = results
        self.result_tree = self._add_lca_results_to_tree(
            results, add_to_distribution, propagate
        )
        self._has_run = True
        return self.result_tree

    def wrapper_data_serializer(self, include_method_units: bool = False):
        method_alias2units: dict[str, str] = {
            method_alias: method_info.bw_method_unit
            for method_alias, method_info in self.experiment.methods.items()
        }

        def data_serializer(data: ScenarioResultNodeData) -> dict:
            result: dict[str, Union[str, float]] = {}
            if data.output:
                result["unit"] = data.output[0] or ""
                result["amount"] = data.output[1] or ""
            if not include_method_units:
                return result | data.results
            else:
                for method_alias, value in data.results.items():
                    final_name = (
                        f"{method_alias} ({method_alias2units[str(method_alias)]})"
                    )
                    result[final_name] = value
                return result

        return data_serializer

    def results_to_csv(
            self,
            file_path: PathLike,
            level_names: Optional[list[str]] = None,
            include_method_units: bool = False,
            warn_no_results: bool = True,
            alternative_hierarchy: BasicTreeNode[ScenarioResultNodeData] = None,
    ):
        """
        Save the results (as tree) to a csv file
         :param file_path:  path to save the results to
         :param level_names: names of the levels to include in the csv
         (must not match length of levels)
         :param alternative_hierarchy: An alternative hierarchy to use for the results,
          which comes from Scenario.rearrange_results.
         :param include_method_units:
        """
        if not self.result_tree:
            raise ValueError(f"Scenario '{self.alias}' has no results")

        if warn_no_results and not self._has_run:
            logger.warning(f"Scenario '{self.alias}' has not been run yet")

        use_tree = self.result_tree
        if alternative_hierarchy:
            use_tree = alternative_hierarchy

        use_tree.to_csv(
            file_path,
            include_data=True,
            level_names=level_names,
            data_serializer=self.wrapper_data_serializer(include_method_units),
        )

    def result_to_dict(
            self,
            include_output: bool = True,
            warn_no_results: bool = True,
            alternative_hierarchy: BasicTreeNode[ScenarioResultNodeData] = None,
    ) -> dict[str, Any]:
        """
        Return the results as a dictionary
        :param include_output:
        :param warn_no_results:
        :param alternative_hierarchy: An alternative hierarchy to use for the results,
        which comes from Scenario.rearrange_results.
        :return:
        """

        def data_serializer(data: ScenarioResultNodeData) -> dict:
            result: dict[str, Any] =  {"results": data.results, "distribution_results": data.distribution_results}
            if include_output:
                result["output"] = {"unit": data.output[0], "amount": data.output[1]}
            if data.bw_activity:
                result["bw_activity"] = data.bw_activity["code"]

            return result

        def recursive_transform(node: BasicTreeNode[ScenarioResultNodeData]) -> dict:
            result: dict[str, Any] = {"alias": node.name, **data_serializer(node._data)}
            if node.children:
                result["children"] = [
                    recursive_transform(child) for child in node.children
                ]
            return result

        if warn_no_results and not self._has_run:
            logger.warning(f"Scenario '{self.alias}' has not been run yet")
        if alternative_hierarchy:
            return recursive_transform(alternative_hierarchy.copy())
        else:
            return recursive_transform(self.result_tree.copy())

    def rearrange_results(
            self, hierarchy: Union[list, dict]
    ) -> BasicTreeNode[ScenarioResultNodeData]:
        alt_result_tree = self.experiment.validate_hierarchy(hierarchy)

        activity_nodes = self.result_tree.iter_leaves()
        alt_activity_nodes = list(alt_result_tree.iter_leaves())
        for node in activity_nodes:
            try:
                alt_node = next(filter(lambda n: n.name == node.name, alt_activity_nodes))
                alt_node._data = node.data
            except StopIteration:
                raise ValueError(
                    f"Activity '{node.name}' not found in alternative hierarchy"
                )
        alt_result_tree.recursive_apply(
            Scenario._recursive_resolve_outputs,
            depth_first=True,
            scenario=self,
            cancel_parents_of=set(),
        )

        alt_result_tree.recursive_apply(
            Scenario._propagate_results_upwards, depth_first=True
        )
        return alt_result_tree

    def __repr__(self):
        return f"<Scenario '{self.alias}'>"



from dataclasses import dataclass, field
from typing import Optional, Union

import bw2data
from bw2data.backends import Activity
from pydantic import Field
from pydantic.dataclasses import dataclass as pydantic_dataclass
from pydantic.v1 import BaseSettings

from enbios2.bw2.util import get_activity
from enbios2.generic.files import PathLike


@pydantic_dataclass
class EcoInventSimpleIndex:
    version: str
    system_model: str


class StrictInputConfig:
    validate_assignment = True
    extra = "forbid"


class OperationConfig:
    validate_assignment = True
    arbitrary_types_allowed = True


@pydantic_dataclass(config=StrictInputConfig)
class ActivityOutput:
    unit: str
    magnitude: float = 1.0


@pydantic_dataclass(config=StrictInputConfig)
class ExperimentActivityId:
    database: Optional[str] = None
    code: Optional[str] = None
    # search and filter
    name: Optional[str] = None
    location: Optional[str] = None
    # additional filter
    unit: Optional[str] = None
    # internal-name
    alias: Optional[str] = None

    def get_bw_activity(
        self, allow_multiple: bool = False
    ) -> Union[Activity, list[Activity]]:
        if self.code:
            if not self.database:
                return get_activity(self.code)
            else:
                return bw2data.Database(self.database).get(self.code)
        elif self.name:
            filters = {}
            if self.location:
                filters["location"] = self.location
                assert (
                    self.database in bw2data.databases
                ), f"database {self.database} not found"
                search_results = bw2data.Database(self.database).search(
                    self.name, filter=filters
                )
            else:
                search_results = bw2data.Database(self.database).search(self.name)
            if self.unit:
                search_results = list(
                    filter(lambda a: a["unit"] == self.unit, search_results)
                )
            assert len(search_results) == 0, (
                f"No results for brightway activity-search:"
                f" {(self.name, self.location, self.unit)}"
            )
            if len(search_results) > 1:
                if allow_multiple:
                    return search_results
                assert False, (
                    f"results : {len(search_results)} for brightway activity-search:"
                    f" {(self.name, self.location, self.unit)}. Results are: "
                    f"{search_results}"
                )
            return search_results[0]
        else:
            raise ValueError("No code or name specified")

    def fill_empty_fields(
        self, fields: Optional[list[Union[str, tuple[str, str]]]] = None, **kwargs
    ):
        if not fields:
            fields = []
        for _field in fields:
            if isinstance(_field, tuple):
                if not getattr(self, _field[0]):
                    setattr(self, _field[0], kwargs[_field[1]])
            else:
                if not getattr(self, _field):
                    setattr(self, _field, kwargs[_field])


@pydantic_dataclass(config=StrictInputConfig)
class SimpleScenarioActivityId:
    name: str
    code: str
    alias: str

    def __hash__(self):
        return hash(self.code)


# this is just for the schema to accept an array.
ExperimentActivityOutputArray = tuple[str, float]

ExperimentActivityOutput = Union[ActivityOutput, ExperimentActivityOutputArray]


@pydantic_dataclass(config=StrictInputConfig)
class ExperimentActivityData:
    """
    This is the dataclass for the activities in the experiment.
    the id, is
    """

    id: ExperimentActivityId = Field(
        ..., description="The identifies (method to find) an activity"
    )
    output: Optional[ExperimentActivityOutput] = Field(
        None, description="The default output of the activity"
    )
    orig_id: Optional[ExperimentActivityId] = Field(
        None, description="Temporary copy of the id"
    )

    @property
    def alias(self):
        return self.id.alias


@pydantic_dataclass(config=OperationConfig)
class ExtendedExperimentActivityData:
    id: ExperimentActivityId
    orig_id: ExperimentActivityId
    output: "ActivityOutput"
    bw_activity: Activity
    default_output_value: Optional[
        float
    ] = 1.0  # this is the value converted to the default (bw) unit

    def __hash__(self):
        return self.bw_activity["code"]

    @property
    def alias(self):
        return self.id.alias


@pydantic_dataclass(config=StrictInputConfig)
class ExperimentMethodData:
    id: tuple[str, ...]
    alias: Optional[str] = None

    @property
    def alias_(self) -> str:
        return str(self.alias)


@pydantic_dataclass(config=StrictInputConfig)
class ExperimentMethodPrepData:
    id: tuple[str, ...]
    alias: str
    bw_method_unit: str


@pydantic_dataclass(config=StrictInputConfig, repr=False)
class ExperimentScenarioData:
    # map from activity id to output. id is either as original (tuple) or alias-dict
    activities: Optional[
        Union[
            list[
                tuple[Union[str, ExperimentActivityId], ExperimentActivityOutput]
            ],  # alias or id to output
            dict[str, ExperimentActivityOutput],
        ]
    ] = None  # alias to output, null means default-output (check exists)

    # either the alias, or the id of any method. not method means running them all
    methods: Optional[
        Union[list[Union[ExperimentMethodData, str]], dict[str, tuple[str, ...]]]
    ] = None
    alias: Optional[str] = None

    @staticmethod
    def alias_factory(index: int):
        return f"Scenario {index}"


@pydantic_dataclass(config=StrictInputConfig)
class ExperimentScenarioPrepData:
    activities: dict[SimpleScenarioActivityId, ExperimentActivityOutput] = Field(
        default_factory=dict
    )
    methods: list[ExperimentMethodData] = Field(default_factory=list)


@pydantic_dataclass(config=StrictInputConfig)
class ExperimentConfig:
    warn_default_demand: bool = True  # todo: bring this back
    include_bw_activity_in_nodes: bool = True
    store_raw_results: bool = False  # store numpy arrays of lca results
    use_k_bw_distributions: int = 1  # number of samples to use for monteCarlo
    store_lca_objects: bool = False
    run_scenarios: Optional[
        list[str]
    ] = None  # list of scenario-alias to run, ALSO AS ENV-VAR
    # only used by ExperimentDataIO
    # base_directory when loading files (activities, methods, ...)
    base_directory: Optional[Union[str, PathLike]] = None
    # those are only used for testing
    debug_test_is_valid: bool = True
    debug_test_replace_bw_config: Union[
        bool, list[str]
    ] = True  # standard replacement, or bw-project, bw-database
    debug_test_expected_error_code: Optional[int] = None
    debug_test_run: Optional[bool] = False
    note: Optional[str] = None


ActivitiesDataRows = list[ExperimentActivityData]
ActivitiesDataTypes = Union[ActivitiesDataRows, dict[str, ExperimentActivityData]]
# with path
ActivitiesDataTypesExt = Union[
    ActivitiesDataRows, dict[str, ExperimentActivityData], PathLike
]

MethodsDataTypes = Union[list[ExperimentMethodData], dict[str, tuple[str, ...]]]
# with path
MethodsDataTypesExt = Union[
    list[ExperimentMethodData], dict[str, tuple[str, ...]], PathLike
]

HierarchyDataTypes = Union[list, dict]
# with path
HierarchyDataTypesExt = Union[list, dict, PathLike]

ScenariosDataTypes = Union[
    list[ExperimentScenarioData], dict[str, ExperimentScenarioData]
]
# with path
ScenariosDataTypesExt = Union[
    list[ExperimentScenarioData], dict[str, ExperimentScenarioData], PathLike
]


@pydantic_dataclass(config=StrictInputConfig)
class ExperimentData:
    """
    This class is used to store the data of an experiment.
    """

    bw_project: Union[str, EcoInventSimpleIndex] = Field(
        ..., description="The brightway project name"
    )
    activities: ActivitiesDataTypesExt = Field(
        ..., description="The activities to be used in the experiment"
    )
    methods: MethodsDataTypesExt = Field(
        ..., description="The impact methods to be used in the experiment"
    )
    bw_default_database: Optional[str] = Field(
        None,
        description="The default database of activities to be used " "in the experiment",
    )
    hierarchy: Optional[HierarchyDataTypesExt] = Field(
        None, description="The activity hierarchy to be used in the experiment"
    )
    scenarios: Optional[ScenariosDataTypesExt] = Field(
        None, description="The scenarios for this experiment"
    )
    config: ExperimentConfig = Field(
        default_factory=ExperimentConfig,
        description="The configuration of this experiment",
    )


@dataclass
class BWCalculationSetup:
    name: str
    inv: list[dict[Activity, float]]
    ia: list[tuple[str, ...]]

    def register(self):
        bw2data.calculation_setups[self.name] = {"inv": self.inv, "ia": self.ia}


@dataclass
class ScenarioResultNodeData:
    output: tuple[Optional[str], Optional[float]] = (None, None)
    results: dict[str, float] = field(default_factory=dict)
    distribution_results: dict[str, list[float]] = field(default_factory=dict)
    bw_activity: Optional[Activity] = None


class Settings(BaseSettings):
    CONFIG_FILE: Optional[str] = None
    RUN_SCENARIOS: Optional[list[str]] = None


Activity_Outputs = dict[SimpleScenarioActivityId, float]



from pathlib import Path
from typing import Optional, Union

import bw2data
import bw2io
from peewee import JOIN

from enbios2.base.databases import init_databases
from enbios2.base.db_models import EcoinventDataset, BWProjectIndex
from enbios2.const import BASE_ECOINVENT_DATASETS_PATH
from enbios2.generic.enbios2_logging import get_logger

logger = get_logger(__name__)


def analyze_directory(
    directory: Optional[Path] = None, store_to_index_file: bool = True
) -> list[EcoinventDataset]:
    """
    Analyzes a directory and returns the dataset descriptors
        for the ecoinvent datasets in the folder.
        They should have been downloaded from the ecoinvent website,
        unzipped and not renamed.
    :param directory:
    :param store_to_index_file:
    :return:
    """
    init_databases()
    if not directory:
        directory = BASE_ECOINVENT_DATASETS_PATH
    indexes: list[EcoinventDataset] = []

    for directory in directory.glob("*"):
        if directory.is_dir() and directory.name.startswith("ecoinvent"):
            parts = directory.name.split("_")
            parts = parts[0].split() + parts[1:]
            version = parts[1]
            system_model = parts[2]
            type_ = parts[-2] if parts[-2] in ["lci", "lcia"] else "default"
            xlsx = parts[-1] == "xlsx"
            index = EcoinventDataset.exising_or_new(
                version=version,
                system_model=system_model,
                type=type_,
                xlsx=xlsx,
                directory=directory,
            )

            indexes.append(index)

    result_indexes: list[EcoinventDataset] = []
    for index in indexes:
        try:
            index.validate()
            result_indexes.append(index)
        except ValueError as e:
            logger.warning(f"directory format not valid: {e}")

    if store_to_index_file:
        for index in result_indexes:
            if EcoinventDataset.exists(index.identity):
                logger.debug(
                    f"Ecoinvent dataset '{index.identity}' "
                    f"already indexed and will not be added"
                )
                continue
            index.save()
            logger.info(f"Added ecoinvent dataset '{index.identity}'")
    return result_indexes


def add_dataset_index(
    version: str, system_model: str, type: str, xlsx: str, directory: str
):
    """
    Add a dataset index
    :param version:
    :param system_model:
    :param type:
    :param xlsx:
    :param directory:
    :return:
    """
    ds = EcoinventDataset(
        version=version,
        system_model=system_model,
        type=type,
        xlsx=xlsx,
        directory=directory,
    )
    if EcoinventDataset.exists(ds.identity):
        logger.debug(
            f"Ecoinvent dataset '{ds.identity}' already indexed and will not be added"
        )
        return
    ds.save()


def get_ecoinvent_dataset_index(
    *,
    version: Optional[Union[str, list[str]]] = None,
    system_model: Optional[Union[str, list[str]]] = None,
    type_: Optional[Union[str, list[str]]] = None,
    xlsx: Optional[bool] = None,
    has_bw_project: Optional[bool] = None,
) -> list[EcoinventDataset]:
    """
    Get the dataset index for the given parameters
    :param version: ecoinvent version, one or multiple
    :param system_model: system model, one or multiple [cut-off, consequential, apos]
    :param type_: type, one or multiple [lci, lcia, default]
    :param xlsx: True if the dataset is in xlsx format
    :return: list of EcoinventDataset
    """
    # build a query for the given parameters
    init_databases()
    query = EcoinventDataset.select()
    if version:
        if isinstance(version, str):
            version = [version]
        query = query.where(EcoinventDataset.version.in_(version))
    if system_model:
        if isinstance(system_model, str):
            system_model = [system_model]
        query = query.where(EcoinventDataset.system_model.in_(system_model))
    if type_:
        if isinstance(type_, str):
            type_ = [type_]
        query = query.where(EcoinventDataset.type.in_(type_))
    if xlsx is not None:
        query = query.where(EcoinventDataset.xlsx == xlsx)
    if has_bw_project is not None:
        query = (
            query.select()
            .join(BWProjectIndex, JOIN.LEFT_OUTER)
            .where(BWProjectIndex.ecoinvent_dataset.is_null(not has_bw_project))
        )
    return list(query)


def is_resolved_database_available(dataset: EcoinventDataset):
    """
    Checks if the resolved database (lci, lcia from excel)
    is available for the given dataset
    :param dataset:
    :return:
    """
    pass


def auto_import(
    eods: EcoinventDataset,
    project_name: Optional[str] = "ecoinvent",
    database_name: Optional[str] = None,
) -> Optional[bw2io.SingleOutputEcospold2Importer]:
    """
    Automatically imports the given ecoinvent dataset into a new project and database.
    Also creates the BWProjectIndex
    YOU SHOULD MAKE SURE THAT BRIGHTWAY DATA CAN DEAL
     WITH THAT PARTICULAR VERSION OF ECOINVENT.

    :param eods: eoinvent dataset (should be indexed)
    :param project_name:
    :param database_name:
    :return:
    """
    if eods.xlsx or not eods.type == "default":
        raise ValueError(
            f"Only default datasets and non xlsx are supported. Passed: {eods}"
        )
    exists = get_ecoinvent_dataset_index(
        version=eods.version,
        system_model=eods.system_model,
        type_=eods.type,
        xlsx=eods.xlsx,
    )
    if not exists:
        eods.save()
    else:
        eods = exists[0]
    if eods.bw_project_index:
        logger.info(f"Already imported and indexed: {eods.bw_project_index}")
        return
    if project_name in bw2data.projects:
        logger.debug(f"Project already exists: {project_name}. switching to it.")
        bw2data.projects.set_current(project_name)
    else:
        logger.debug(f"Creating new project. {project_name}")
        bw2data.projects.create_project(project_name)
        bw2data.projects.set_current(project_name)
        bw2io.bw2setup()

    if not database_name:
        database_name = eods.identity
    if database_name in bw2data.databases:
        raise ValueError(f"Database already exists: {database_name}")
    logger.info(f"Importing ecoinvent dataset to {project_name}/{database_name}")
    importer = bw2io.SingleOutputEcospold2Importer(
        eods.dataset_path.as_posix(), database_name
    )
    importer.apply_strategies()
    importer.statistics()
    if importer.statistics()[2] == 0:
        importer.write_database()
        BWProjectIndex.create(
            project_name=project_name,
            database_name=database_name,
            ecoinvent_dataset=eods,
        )
    else:
        print(
            "There are unlinked exchanges. Database will not be written. "
            "Method returns importer "
            "(you can inspect, manipulable and write it manually)."
        )
    return importer


def analyse_and_import():
    """
    analyse the ecoinvent directory and import all datasets
    """

    analyze_directory(store_to_index_file=True)
    indexes = get_ecoinvent_dataset_index()
    for index in indexes:
        if index.version.startswith("3.9"):
            try:
                auto_import(index)
            except ValueError as e:
                logger.warning(e)


if __name__ == "__main__":
    analyse_and_import()

    # bw2data.projects.delete_project("ecoinvent", True)

    # print(list(get_ecoinvent_dataset_index(xlsx=True)))
    # print(list(get_ecoinvent_dataset_index(xlsx=True))[0].dataset_path)
